{-------------------------- Dedication ---------------------------------

    Dedicated to our daughter Brianna Grieve stillborn 6th July 1996

 -----------------------------------------------------------------------}

{ TxyGraph version 3.01
  see documentation (default in /docs/index.htm) for
              - current known buglist
              - how to use TxyGraph}

{ copyright  Kestral Computing
  contact grahameg@kestral.com.au }

{ definitions:}
{$I xyDefine.pas}

{--------------- implementation section heading labels ---------------

    #1a. TSeries - administration
    #1b. TSeries - property servers
    #1c. TSeries - drawing routines
    #2a. TAxis  - scaling routines
    #2b. TAxis - property servers
    #3.  TDimensions
    #4.  TAppearance
    #5.  TLegend
    #6a. TxyGraph  - administration
    #6b. TxyGraph  - mouse routines
    #6c. TxyGraph  - metrics routines
    #6d. TxyGraph  - painting routines
    #6e. TxyGraph  - property servers
    #6f. TxyGraph  - series routines
    #6g. TxyGraph  - data manipulation
    #6h. TxyGraph  - Marks routines
    #6i. TxyGraph  - External interaction
                       (clipboard, assign, savetofile etc)

 ----------------------------------------------------------------------}

unit xyGraph;

{$B-}

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Forms, ExtCtrls,
  Controls, Printers, menus, xyData;

const
  versionstamp = '3.01';
  xyGraphFileSig = 'xygg';
  xyGraphFileVersion = '0';
  defMinYScale = 0.0;
  defMinXScale = 0.0;
  drawBefore = false;
  drawAfter = true;
  AllowedLogTickCounts = [0,1,2,5,7,9];
    {InitLogTicks and AdjustLogTickCount must correspond.
     AdjustLogLabelDecs must correspond with the result of InitLogTicks.}
  MaxLogTickCount = 9; {Must equal largest value in AllowedLogTickCounts}

  MouseToleranceMove = 4; { pixels mouse must move before it counts }
  MouseToleranceHit = 5; { how far from target mouse is allowed to be }

var
  CF_xyGraphFmt:word;

type
  et_PointShape = (ps_Square, ps_Circle, ps_Diamond, ps_Cross, ps_Wide);
  et_BoundType = (bt_None, bt_AsSet, bt_1SD, bt_2SD, bt_3SD, bt_allSD);
  et_MarkType = (mtXMark,mtYMark,mtPoint,mtLine);
  TLegendStyle = (lsAcross, lsDown);
  TLegendStatus = (lsNone, lsNameOnly, lsPoints, lsLines, lsAll);

  pMark = ^Tmark;
  tMark = record
          id:integer;
          x1,y1,x2,y2:double;
          color:Tcolor;
          caption:String;
          next:pmark;
          marktype:et_Marktype;
          status:boolean;
          end;

  PlotFunction = function(x: Double; var parms): Double;

  Dateticks = (dt_minute, dt_5minute, dt_halfhourly, dt_hourly,
         dt_daily, dt_weekly, dt_monthly, dt_2monthly, dt_quarterly,
         dt_annually, dt_decade, dt_century, dt_custom);

  TLogTickInfo = record
         LogTickArray: array[0..MaxLogTickCount+1] of Double;
         LogTickCount: Word;
         LogTickIndex: Word;
         LogStepping: Boolean; {signals linear steps}
         DrawMinorLabels: Boolean;
         end;

  TSeriesEvent =  procedure (sender:TObject; index:integer) of object;
  TPaintEvent = procedure (sender:TObject; Canvas:TCanvas) of object;

  TxyGraph = class;
  TAxis = class;

{------------------------ TSeries --------------------------------*}
  TSeries = class (TDataSeries)
  private
    next: Tseries;
    FGraph: TxyGraph;
    FWhichYAxis: TAxis;
    FActive, FDrawLine, FDrawPoints, FFillPoints,
    FHoldUpdates, UpdatesWaiting: Boolean;
    FLineColor, FPointColor, FBoundsColor: TColor;
    FPointShape: et_PointShape;
    FLineStyle,FBoundsLineStyle: TPenStyle;
    FSeriesIndex:longint;
    FPointSize: Word;
    FUpperBound,FLowerBound:Double;
    FBoundsType: et_BoundType;
    FLegendStatus: TLegendStatus;
    FRegrColor: TColor;
    FRegrLineStyle: TPenStyle;

    procedure setHoldUpdates(v: Boolean);
    procedure setActive(v: Boolean);
    procedure setDrawLine(v: Boolean);
    procedure SetDrawPoints(v: Boolean);
    procedure setFillPoints(v: Boolean);
    procedure SetLineColor(v: TColor);
    procedure SetPointColor(v: TColor);
    procedure SetBoundsColor(v: TColor);
    procedure SetPointShape(v: et_pointshape);
    procedure setLinestyle(v: TPenStyle);
    procedure setPointSize(v: Word);
    procedure SetWhichYAxis(v: TAxis);
    procedure SetUpperBound(v:double);
    procedure SetLowerBound(v:Double);
    procedure setBoundsType(v:et_BoundType);
    procedure SetBoundsLineStyle(v: TPenstyle);
    procedure SetRegrColor(v: TColor);
    procedure SetRegrLineStyle(v: TPenstyle);
    procedure SetLegendStatus(v:TlegendStatus);
  protected
    procedure RequestPaint(TheMessage:TDSChangeType); override;
    procedure WarnTooManyPoints; override;
    procedure DrawMyLine;
    procedure DrawMyPoints;
    procedure DrawBounds;
   {$IFDEF STATISTICS}
    procedure DrawMyRegression;
   {$ENDIF}
  public
    constructor Create(aGraph:TxyGraph; i:integer);
    destructor Destroy;  override;
    procedure Draw;

    procedure refreshdata; override;

    property SeriesIndex: Longint read FSeriesIndex;
    property NextSeries: TSeries read Next;
    property graph: TXyGraph read FGraph;

    property HoldUpdates: Boolean read FHoldUpdates write setHoldUpdates;
    property Active: Boolean read FActive write setActive stored true;
    property DrawLine: Boolean read FDrawLine write setDrawLine stored true;
    property DrawPoints: Boolean read FDrawPoints write SetDrawPoints stored true;
    property FillPoints: Boolean read FFillPoints write setFillPoints stored true;
    property LineColor: TColor read FLineColor write SetLineColor stored true;
    property PointColor: TColor read FPointColor write SetPointColor stored true;
    property BoundsColor: TColor read FBoundsColor write SetBoundsColor stored true;
    property PointShape: et_PointShape read FPointShape write SetPointShape stored true;
    property LineStyle: TPenStyle read FLineStyle write setLinestyle stored true;
    property PointSize: Word read FPointSize write setPointsize stored true;
    property WhichYAxis: TAxis read FWhichYAxis write SetWhichYAxis stored true;
    property UpperBound:double read FUpperBound write SetUpperBound stored true;
    property LowerBound:double read FLowerBound write SetLowerBound stored true;
    property BoundsType: et_BoundType read FBoundstype write SetBoundstype stored true;
    property BoundsLineStyle: TPenStyle read FBoundsLineStyle write SetBoundsLineStyle stored true;
    property RegressionLineColor: TColor read FRegrColor write SetRegrColor stored true;
    property RegressionLineStyle: TPenStyle read FRegrLineStyle write SetRegrLineStyle stored true;
    property LegendStatus:TLegendStatus read FLegendStatus write SetLegendStatus stored true;
  end {class TSeries};

  THookType = (ht_PlotData {$IFDEF STATISTICS}, ht_PlotResiduals {$ENDIF});
  pHookedSeries = ^THookedSeries;
  THookedSeries = record
      id:TSeries;
      hooktype:THookType;
      Series:integer;
      next: pHookedSeries;
    end;


{------------ TAxis, TDimensions, TAppearance ----------------}

  TAxisOffsetType = (ao_Minimum, ao_Maximum, ao_percentage, ao_absolute);

  TAxis = class (TPersistent)
  private
    FGraph: TxyGraph;
    SecondAxis, ScaledOK: boolean;
    FM: Double;
    oMin, oMax: Double;
    FMin, FMax, FStep: Double;
    FAutoSizing: Boolean;
    FAutoStepping: Boolean;
    FLogging: Boolean;
    FLogTickInfo: TLogTickInfo;
    FAxisTitle: String;
    FLabelDec: Integer;
    FMinScale, FOffset: Double;
    FShowAsTime, FShowAxis, FReversed, FGridlines, FAutoLabelDecimals: Boolean;
    FDateFormat: string;
    FDateTickType: dateticks;
    FOffsetType:TAxisOffsetType;
    procedure SetLabelDec(v: Integer);
    function  GetLogTickCount: Word;
    procedure SetLogTickCount(v: Word);
    procedure SetLogging(v: Boolean);
    procedure SetAutoSizing(v: Boolean);
    procedure SetAutoStepping(v: Boolean);
    procedure SetMax(v: Double);
    procedure SetMin(v: Double);
    procedure SetStep(v: Double);
    procedure SetMinDiff(v: Double);
    procedure SetAxisTitle(v: String);
    procedure setShowAsTime(v: Boolean);
    procedure setDateFormat(v: string);
    procedure setDateTickType(v:dateticks);
    procedure setshowAxis(v:boolean);
    procedure setreversed(v:boolean);
    procedure SetOffsetType(v:TAxisOffsetType);
    procedure setoffset(v:double);
    procedure SetGridLines(v: Boolean);
  protected
    procedure InitLogTicks;
    procedure AdjustLogTickCount;
    procedure CheckDrawMinorLabels;
    function  GetStep: Double;
    function  getDateStep: double;
    procedure SetMinMax;
    procedure SetDateMinMax;
    procedure SetLogMinMax;
    function  DoResize: Boolean;
    function  GetFirstTick(var logTickInfo: TLogTickInfo): Double;
    function  GetFirstDateTick:double;
    function  GetNextTick(tick: Double; var logTickInfo: TLogTickInfo;
                          var drawThisLabel: Boolean): Double;
    function  GetNextDateTick(tick: Double):double;
    procedure AdjustLabelDecs;
    procedure AdjustLogLabelDecs(v: Double);
    procedure CalcMetrics;
    function  LabelString(tick: Double): String;
  public
    function  CheckScale: Boolean;
  published
    property Title: string read FAxisTitle write SetAxisTitle;
    property LabelDecimals: Integer read FLabelDec write SetLabelDec default 1;
    property AutoLabelDecimals: boolean read FAutoLabelDecimals write FAutoLabelDecimals default true;
    property LogCycleDivisions: Word read GetLogTickCount write SetLogTickCount default 2;
    property LogScale: Boolean read FLogging write SetLogging;
    property Max: Double read FMax write SetMax;
    property Min: Double read FMin write SetMin;
    property StepSize: Double read FStep write SetStep;
    property MinScaleLength: Double read FMinScale write SetMinDiff;
    property ShowAsTime: Boolean read FShowASTime write setShowAsTime default false;
    property DateTimeFormat: string read FDateFormat write SetDateFormat;
    property DateTickType: dateticks read FDateTickType write setDateTickType;
    property ShowAxis: boolean read FShowAxis write SetShowAxis;
    property Reversed: boolean read FReversed write setreversed;
    property OffsetType:TAxisOffsetType read FOffsettype write SetOffsetType;
    property Offset:double read FOffset write setoffset;
    property Gridlines: boolean read FGridlines write SetGridlines;

   { these 2 properties must come last to override the other properties'
      effects on the values at load time: }
    property AutoSizing: Boolean read FAutoSizing write SetAutoSizing;
    property AutoStepping: Boolean read FAutoStepping write SetAutoStepping;
  end {class TAxis};

  TDimensions = class (TPersistent)
  private
    FGraph: TxyGraph;
    FLeft, FRight, FTop, FBottom: Word;
    FTMLength: Word;
    FXAxisTitleDistance: Integer;
    FYAxisTitleDistance: Integer;
    FXAxisLabelDistance: Integer;
    FYAxisLabelDistance: Integer;
    FGraphTitleDistance: Integer;
    FScalePct: Word;
    FXOffsetPct, FYOffsetPct: Word;
    procedure SetMargBottom(v: Word);
    procedure SetMargTop(v: Word);
    procedure SetMargLeft(v: Word);
    procedure SetMargRight(v: Word);
    procedure SetTickLength(v: Word);
    procedure SetXAxisTitleDistance(v: Integer);
    procedure SetYAxisTitleDistance(v: Integer);
    procedure SetXAxisLabelDistance(v: Integer);
    procedure SetYAxisLabelDistance(v: Integer);
    procedure SetGraphTitleDistance(v: Integer);
    procedure SetScale(v: Word);
    procedure SetXOffset(v: Word);
    procedure SetYOffset(v: Word);
  published
    property BottomMargin: Word read FBottom write SetMargBottom default 40;
    property LeftMargin: Word read FLeft write SetMargLeft default 40;
    property RightMargin: Word read FRight write SetMargRight default 15;
    property TopMargin: Word read FTop write SetMargTop default 30;
    property TickLength: Word read FTMLength write SetTickLength default 4;
    property XAxisTitleOffset: Integer read FXAxisTitleDistance write SetXAxisTitleDistance default 4;
    property XAxisLabelOffset: Integer read FXAxisLabelDistance write SetXAxisLabelDistance default 2;
    property YAxisTitleOffset: Integer read FYAxisTitleDistance write SetYAxisTitleDistance default 4;
    property YAxisLabelOffset: Integer read FYAxisLabelDistance write SetYAxisLabelDistance default 2;
    property GraphTitleOffset: Integer read FGraphTitleDistance write SetGraphTitleDistance default 7;
   {set print offsets as integer percent (<=100) of page width, height}
    property PrintXOffsetPct: Word read FXOffsetPct write SetXOffset default 5;
    property PrintYOffsetPct: Word read FYOffsetPct write SetYOffset default 20;
   {set print scale as integer percent (<=100) of full page}
    property PrintScalePct: Word read FScalePct write SetScale default 90;
  end {TDimensions};

  TAppearance = class (TPersistent)
  private
    FGraph: TxyGraph;
    FTickMarks, FShowMarks, FCrossAtZero,FPrintLineStyle: Boolean;
    FAllowedOut, FAllowDuplicates: Boolean;
    FLabelGraph: Boolean;
    FGridColor, FAxesColor, FBkgdColor, FMarginColor, FCrossColor: TColor;
    FGridStyle: TPenStyle;
    FBkgdWhenPrint:Boolean;
    FCaption: string;
    FErrorCaption: string;
    FGraphTitle: string;
    FMinSteps,FMaxSteps, FCrossLength: Word;
    FTitleFont, FCaptionFont, FLabelFont: TFont;
    procedure SetAxesColor(v: TColor);
    procedure SetBkgdColor(v: TColor);
    procedure SetMarginColor(v: TColor);
    procedure SetErrorCaption(v: string);
    procedure SetGridColor(v: TColor);
    procedure SetGridStyle(v: TPenStyle);
    procedure SetLabelGraph(v: Boolean);
    procedure SetAllowedOut(v: Boolean);
    procedure SetShowMarks(v: Boolean);
    procedure SetTickMarks(v: Boolean);
    procedure SetGraphTitle(v: string);
    procedure SetMinSteps(v: Word);
    procedure SetMaxSteps(v: Word);
    procedure setTitleFont(v:TFont);
    procedure setLabelFont(v:TFont);
    procedure setCaptionFont(v:TFont);
    procedure setCrossAtZero(v:boolean);
    procedure setCrossColor(v:tColor);
    procedure setCrosslength(v:word);
    procedure SetAllowDuplicates(v: Boolean);
    procedure FontChanged(Sender: TObject);
 published
    property AllowDuplicates: Boolean read FAllowDuplicates write SetAllowDuplicates default false;
    property AxesColor: TColor read FAxesColor write SetAxescolor default clBlack;
    property BackgroundColor: TColor read FBkgdColor write SetBkgdColor;
    property MarginColor: TColor read FMarginColor write SetMarginColor;
    property PrintBkgndColor: Boolean read FBkgdWhenPrint write FBkgdWhenPrint;
    property ErrorCaption: string read FErrorCaption write SetErrorCaption;
    property GridColor: TColor read FGridColor write SetGridcolor default clSilver;
    property GridStyle: TPenStyle read FGridStyle write SetGridStyle default psDot;
    property ShowGraphLabels: Boolean read FLabelGraph write SetLabelGraph default True;
    property PlotOffGraph: Boolean read FAllowedOut write SetAllowedOut default False;
    property ShowMarks: Boolean read FShowMarks write SetShowMarks default false;
    property ShowTicks: Boolean read FTickMarks write SetTickmarks default True;
    property GraphTitle: string read FGraphTitle write SetGraphTitle;
    property MinSteps: Word read FMinSteps write SetMinSteps default 5;
    property MaxSteps: Word read FMaxSteps write SetMaxSteps default 50;
    property TitleFont: TFont read FTitleFont write setTitleFont;
    property CaptionFont: TFont read FCaptionFont write setCaptionFont;
    property LabelFont: TFont read FLabelFont write setLabelFont;
    property CrossAtZero:Boolean read FCrossatZero write setCrossAtZero default false;
    property CrossColor:tColor read FCrossColor write SetCrossColor default clRed;
    property Crosslength:word read FCrosslength write SetCrossLength default 4;
    property PrintLineStyle:boolean read FPrintLineStyle write FPrintLineStyle default true;
  end {TAppearance};

  TLegend = class (TPersistent)
  private
     FGraph: TxyGraph;
     Fvisible:Boolean;
     Fcolor:TColor;
     FborderStyle:TBorderStyle;
     FLayout:TLegendStyle;
     Ftop, Fleft, Fwidth, Fheight, FSymbolSpace, FXMargin, FYmargin: integer;
     Ffont:TFont;
     procedure setvisible(v:Boolean);
     procedure setcolor(v:TColor);
     procedure setborderStyle(v:TBorderStyle);
     procedure setLayout(v:TLegendStyle);
     procedure settop(v:integer);
     procedure setleft(v:integer);
     procedure setwidth(v:integer);
     procedure setheight(v:integer);
     procedure setfont(v:TFont);
     procedure setSymbolSpace(v:integer);
     procedure setXMargin(v:integer);
     procedure setYMargin(v:integer);

     procedure fontchanged(sender: TObject);
  published
     property visible:Boolean read FVisible write SetVisible default false;
     property color:TColor read Fcolor write SetColor default clWhite;
     property borderStyle:TBorderStyle read FBorderstyle write SetBorderStyle default bsNone;
     property Layout:TLegendStyle read FLayout write SetLayout default lsAcross;
     property top:integer read FTop write settop default 10;
     property left:integer read FLeft write setleft default 10;
     property width:integer read FWidth write setWidth;
     property height:integer read Fheight write setheight default 20;
     property font:TFont read FFont write setFont;
     property SymbolSpace:integer read FSymbolSpace write setSymbolSpace default 20;
     property XMargin:integer read FXMargin write setXmargin default 10;
     property YMargin:integer read FYMargin write setYMargin default 5;
  end;

{-------------------------- TxyGraph ------------------------------*}

  TxyGraph = class(TCustomPanel)
  private
    { Private declarations }
    FCanvas, FAltCanvas: TCanvas;   {FAltCanvas always points a full
        FCanvas. this is important when a metafile is being drawn, as some
        information, such as textwidth is not available from FCanvas (D1.0)
          i.e. when painting,   FCanvas = FAltCanvas = Canvas
               when printing,   FCanvas = FAltCanvas = printer.canvas
               when metafiling, FCanvas = Metafile.canvas,
                   and FAltCanvas := Canvas  }
    FSeries: TSeries;
    FDimensions: TDimensions;
    FAppearance: TAppearance;
    FXAxis, FYAxis, FYAxis2: TAxis;
    FLegend:TLegend;
    FLastSeries:integer;
    FMarks: pMark;
    AllScalesOK, FNoGridlines: Boolean;
    FHookedSeries: pHookedSeries;

  {Graph property data:}
    FPlotting, FPainting: Boolean;
    FScale: Double;
    FXOffset, FYOffset, FCurrWidth, FCurrHeight: Word;


   {dragging: }
    dRect: tRect;
    dx, dy: Integer;
    FDragging, FLegDragging, FHasDragged: Boolean;
    FDraggable: Boolean;
    FHoldXAS, FHoldYAS, FHoldY2AS: boolean;
    FHoldXmin, FHoldXmax, FHoldYmin, FHoldYmax, FHoldY2min, FHoldY2max: double;

   {Events:}
    FOnNewSeries, FOnLogError, FOnTooManyPoints, FOnDeleteSeries: TSeriesEvent;
    FOnRescale, FOnScaleError, FOnWantDesigner: TNotifyEvent;
    FOnPaintEnd: TPaintEvent;


    FFilingSeriesCount,FFilingMarksCount:word;
    FSaveData:boolean;

    procedure SetPlotting(v: Boolean);
    procedure SetHasDragged(v:boolean);

   {These two are needed for Series[] property:}
    function GetSeriesInt(i:integer):TSeries;
    function CreateNewSeries(i:integer):TSeries;

   {initialise all properties. Run on create and before
    readfromfile}
    procedure setDefaultPropertyValues;

   {routines concerned with writing a full image of the graph to
    the component image}
    procedure defineproperties(Filer: TFiler); override;
    procedure ReadSeriesCount(Reader: TReader);
    procedure writeseriescount(writer: TWriter);
    procedure WriteAllSeries(Stream: TStream);
    procedure ReadAllSeries(Stream: TStream);
    procedure writeMarkCount(writer: TWriter);
    procedure ReadMarkCount(Reader: TReader);
    procedure WriteAllMarks(Stream: TStream);
    procedure ReadAllMarks(Stream: TStream);

  protected
    procedure DoTightFit;
    procedure ClipGraph;
    procedure UnclipGraph;
    procedure CalcMetrics;
    function  fx(v: Double): Integer;
    function  fy(v: Double): Integer;
    function  fy2(v: Double): Integer;
    procedure DrawLineSegment(x1, y1, x2, y2: Integer);
    procedure DrawXGridlines;
    procedure DrawYGridlines(CYAxis: TAxis);
    procedure DrawXTickMarks(base:integer; reverse:boolean);
    procedure DrawYTickMarks(CYAxis: TAxis; base:integer; reverse:boolean);
    function  DrawXLabels(base:integer; reverse:boolean):integer;
    function  DrawYLabels(CYAxis: TAxis; base:integer; reverse:boolean):integer;
    procedure DrawGraphTitle;
    procedure DrawMarks(when: Boolean);
    procedure DrawXAxis;
    procedure DrawYAxis;
    procedure DrawY2Axis;
    procedure DrawAxes;
    procedure ColorBackground;
    procedure PaintCross;
    procedure DrawLegend;
    procedure PaintGraph;
    procedure Paint; override;
    function  CheckScalesOK: Boolean;
    procedure DoRescaleEvent;
    function  DataFound(whichaxis:TAxis): Boolean;

   {Implements dragging and resizing:}
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure SetUpDragging(X,Y: Integer);
    procedure MouseMove(Shift: TShiftState; X,Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X,Y: Integer); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy; override;

    property checkScales:boolean read AllScalesOK;

   { data routines }
    property serieshandle:TSeries read FSeries; {provided to iterate through
                                                 existing series}
    property LastSeriesIndex:integer read FLastSeries;
    property  Series[i:integer]: TSeries read GetSeriesInt; default;
    function  FindSeries(i:integer): Boolean;
    procedure ClearAll;

   { hooking other graph series }
    procedure Hookseries(Source:TSeries; Ahooktype:THooktype;
                     Dest:Integer; WantPersist:boolean);
    procedure unhookseries(source:TSeries; WantPersist:boolean);
    procedure HookedSeriesEvent(Sender:TObject; TheMessage:TDSChangeType);

   { Marks }
    procedure ClearMarks;
    procedure Addmark(id: Integer; x1,y1, x2,y2: Double;
                   c: Tcolor; name: string;
                   marktype:et_marktype; status:boolean);
    function  DeleteMark(id: Integer): Boolean;

   { for mouse events }
    property HasDragged: boolean read FHasDragged write SetHasDragged;
    function  GetMouseXY(sN, x, y: Integer; var index: Integer;
                   var xScaled, yScaled,
                   xNearest, yNearest: Double): Boolean;
    function  GetMouseX(x: Integer): Double;
    function  GetMouseY(y: Integer): Double;
    function  GetMouseY2(y: Integer): Double;
    function  GetMouseVals(sN, x: Integer; var index: Integer;
                   var xNearest, yNearest: Double): Boolean;
    function findNearestPoint({var }series, x,y:integer):pgPoint;

   { PrintOnPage allows more than one graph to be printed on a page; calls
    to it must be preceded by BeginDoc amd followed by EndDoc.
    Print prints only the current graph on its own sheet.
    The graph is printed without change in aspect ratio.
    The size of the printed graph is controlled by PrintScalePct,
    which makes the dimension of the printed graph a percentage of the page
    dimension (on the axis with the larger Screen/Page ratio)
    The location of the upper left corner of the graph is controlled by
    PrintX(Y)OffsetPct, which locate the corner as percentages of the
    page height and width}
    procedure PrintOnPage;
    procedure Print;

    procedure FillMetafile(wmf:TMetafile);
    procedure savetoWMF(const fname:string);
    procedure FillBitmap(bmp:TBitmap);
    procedure SaveToBitmap(const fname:string);
    procedure CopyToClipboard;
    procedure PasteFromClipboard;

  { A convenient way to graph a function; automatically fills graph; handles
    singularities in the function.
    F must be far; parms may be of any type (usu. array of Double) -- no error checks;
    the graph of the function will be drawn from x=x1 to x=x2;
    if x1=x2, will be set to XMin, XMax; if larger, will be clipped to XMin, XMax
    if steps=0, will choose "enough" to be smooth }
    procedure DrawFunction(F: PlotFunction; var parms; x1, x2: Double;
                   color: TColor; style: TPenStyle; steps: Word);

    procedure savetofile(fname:string; savedata:boolean);
    procedure savetostream(f:TStream; savedata:boolean);
    procedure readfromfile(fname:string);
    procedure readfromStream(f:TStream);
    procedure Assign(source:TPersistent); override;

  published
    property DragAllowed: Boolean read FDraggable write FDraggable default true;

    property XAxis: TAxis read FXAxis write FXAxis;
    property YAxis: TAxis read FYAxis write FYAxis;
    property YAxis_Second: TAxis read FYAxis2 write FYAxis2;
    property Dimensions: TDimensions read FDimensions write FDimensions;
    property Appearance: TAppearance read FAppearance write FAppearance;
    property Legend: TLegend read FLegend write Flegend;

    property OnLogError: TSeriesEvent read FOnLogError write FOnLogError;
    property OnNewSeries: TSeriesEvent read FOnNewSeries write FOnNewSeries;
    property OnDeleteSeries: TSeriesEvent read FOnDeleteSeries write FOnDeleteSeries;
    property OnTooManyPoints: TSeriesEvent read FOnTooManyPoints write FOnTooManyPoints;
    property OnRescale: TNotifyEvent read FOnRescale write FOnRescale;
    property OnScaleError: TNotifyEvent read FOnScaleError write FOnScaleError;
    property OnPaintEnd: TPaintEvent read FOnPaintEnd write FOnPaintEnd;
    property OnWantDesigner: TNotifyEvent read FOnWantDesigner write FOnWantDesigner;

    property Align;
    property BevelInner default bvNone;
    property BevelOuter default bvNone;
    property BevelWidth;
    property BorderStyle default bsSingle;
    property BorderWidth;
    property Ctl3D;
    property Cursor;
    property ParentCtl3D;
    property ParentFont;
    property PopupMenu;
    property Visible;
    property Plotting: Boolean read FPlotting write SetPlotting;

    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnResize;
  end {class TxyGraph};

implementation

uses
  clipbrd, xyFile {$IFNDEF WIN32}, xyMeta {$ENDIF};

{----------------------------------------------------------------------------
    #1a. TSeries - administration
 ----------------------------------------------------------------------------}

 constructor TSeries.create(aGraph:TxyGraph; i:integer);
 begin
   inherited create;
   next := nil;
   FGraph := aGraph;
   WhichYAxis := FGraph.FYAxis;
   FSeriesIndex := i;
   FseriesName := 'Series ' +inttostr(i);
   if FGraph.FLastSeries < i then FGraph.FLastSeries := i;
   FActive := true;
   FDrawLine := true;
   FDrawPoints := true;
   FHoldUpdates := false;
   FAllowDuplicates := FGraph.FAppearance.allowduplicates;
   case FSeriesIndex of
    1:FPointColor := clgreen;
    2:FPointColor := clblue;
    3:FPointColor := clred;
    4:FPointColor := clGray;
    5:FPointColor := clMaroon;
    else FPointColor := clgreen;
   end;
   FLineColor := FPointColor;
   case FSeriesIndex of
    1:FPointShape := ps_Square;
    2:FPointShape := ps_Circle;
    3:FPointShape := ps_Diamond;
    4:FPointShape := ps_Cross;
    5:FPointShape := ps_Square;
    else FPointShape := ps_Circle;
   end;
   FFillPoints := (FSeriesIndex <= 4);

   case FSeriesIndex of
    1:FLineStyle := psSolid;
    2:FLineStyle := psDash;
    3:FLineStyle := psDot;
    4:FLineStyle := PsDashDot;
    5:FLineStyle := PsDashDotDot;
    else FLineStyle := psSolid;
   end;

   FPointSize := 4;
   FUpperBound := 0;
   FLowerBound := 0;
   FBoundsType := bt_none;
   FBoundsLineStyle := psDashDot;
   FRegrColor := clgray;
   FRegrLineStyle := psDot;
   FlegendStatus := lsNone;
end;

destructor TSeries.destroy;
var p1,p2:TSeries;
begin
  {splice self out of FGraph's series list}
  p1 := FGraph.Fseries;
  p2 := nil;
  while p1.Fseriesindex <> Fseriesindex do
    begin
    p2 := p1;
    p1 := p1.next;
    end;
  if p2 = nil then
   FGraph.Fseries := p1.next
  else p2.next := p1.next;
  if assigned(FGraph.FOnDeleteSeries) then
    FGraph.FOnDeleteSeries(self,Fseriesindex);
  inherited destroy;
end;

{----------------------------------------------------------------------------
    #1b. TSeries - property servers
 ----------------------------------------------------------------------------}

procedure TSeries.setHoldUpdates(v:boolean);
begin
 if v <> FHoldUpdates then
   begin
   FHoldUpdates := v;
   if FHoldUpdates then
     updateswaiting := false
   else
     if updateswaiting then RequestPaint(dsClearUpdates);
   end;
end;

procedure TSeries.setActive(v:boolean);
begin
  if v <> FActive then
  begin
    FActive := v;
    RequestPaint(dsChangeLook);
  end;
end;

procedure TSeries.setDrawLine(v:boolean);
begin
  if v <> FDrawline then
  begin
    FDrawline := v;
    RequestPaint(dsChangeLook);
  end;
end;

procedure TSeries.SetDrawPoints(v:boolean);
begin
  if v <> FDrawpoints then
  begin
    FDrawpoints := v;
    RequestPaint(dsChangeLook);
  end;
end;

procedure TSeries.setFillPoints(v:boolean);
begin
  if v <> Ffillpoints then
  begin
    Ffillpoints := v;
    RequestPaint(dsChangeLook);
  end;
end;

procedure TSeries.SetLineColor(v:TColor);
begin
  if v <> FLineColor then
  begin
    FLineColor := v;
    RequestPaint(dsChangeLook);
  end;
end;

procedure TSeries.SetPointColor(v:TColor);
begin
  if v <> FPointColor then
  begin
    FPointColor := v;
    RequestPaint(dsChangeLook);
  end;
end;

procedure TSeries.SetBoundsColor(v:TColor);
begin
  if v <> FBoundsColor then
  begin
    FBoundsColor := v;
    RequestPaint(dsChangeLook);
  end;
end;

procedure TSeries.SetPointShape(v:et_pointshape);
begin
  if v <> FPointShape then
  begin
    FPointShape := v;
    if (FPointShape <> ps_Wide) then
    begin
     if fPointsize < 2 then fPointsize := 2;
     if odd(fPointsize) then Inc(fPointsize);
    end;
    RequestPaint(dsChangeLook);
  end;
end;

procedure TSeries.SetLinestyle(v:TPenStyle);
begin
  if v <> fLinestyle then
  begin
    fLinestyle := v;
    RequestPaint(dsChangeLook);
  end;
end;

procedure TSeries.SetBoundsLinestyle(v:TPenStyle);
begin
  if v <> fBoundsLinestyle then
  begin
    fBoundsLinestyle := v;
    RequestPaint(dsChangeLook);
  end;
end;

procedure TSeries.SetPointsize(v:word);
begin
  if v <> fPointsize then
  begin
    fPointsize := v;
    if (FPointShape <> ps_Wide) then
    begin
     if fPointsize < 2 then fPointsize := 2;
     if odd(fPointsize) then Inc(fPointsize);
    end;
    RequestPaint(dsChangeLook);
  end;
end;

procedure TSeries.SetUpperBound(v:double);
begin
 if v <> FUpperBound then
   begin
   FUpperBound := v;
   if FBoundsType = bt_AsSet then requestpaint(dsChangeLook);
   end;
end;

procedure TSeries.SetLowerBound(v:Double);
begin
 if v <> FLowerBound then
   begin
   FLowerBound := v;
   if FBoundsType = bt_AsSet then requestpaint(dsChangeLook);
   end;
end;

procedure TSeries.setBoundsType(v:et_BoundType);
begin
  if v <> FBoundsType then
    begin
    FBoundsType := v;
    requestpaint(dsChangeLook);
    end;
end;

procedure TSeries.SetRegrColor(v:TColor);
begin
  if v <> FRegrColor then
  begin
    FRegrColor := v;
{$IFDEF STATISTICS}
    if FRegType <> rg_none then RequestPaint(dsChangeLook);
{$ENDIF}
  end;
end;

procedure TSeries.SetRegrLineStyle(v:TPenstyle);
begin
  if v <> FRegrLineStyle then
  begin
    FRegrLineStyle := v;
{$IFDEF STATISTICS}
    if FRegType <> rg_none then RequestPaint(dsChangeLook);
{$ENDIF}
  end;
end;

procedure TSeries.SetLegendStatus(v:TlegendStatus);
begin
  if v <> FLegendStatus then
    begin
    FLegendStatus := v;
    RequestPaint(dsChangeLook);
    end;
end;

procedure TSeries.SetWhichYAxis;
begin
  if v <> FWhichYAxis then
    begin
    FWhichYAxis := v;
    RequestPaint(dsChangeLook);
    end;
end;

procedure TSeries.WarnTooManyPoints;
begin
  if assigned(FGraph.FOnTooManyPoints) then
    FGraph.FOnTooManyPoints(FGraph, FSeriesIndex);
  inherited WarnToomanyPoints;
     { I don't know whether this inherited bit is a good idea or not - Don't
       think it's very likely to be used. But when adding a lot of points,
       the overhead is insignificant, in that it's expected to be slow?? }
end;

procedure TSeries.refreshdata;
begin
  HoldUpdates := true;
  try
    inherited refreshdata;
  finally
    HoldUpdates := false;
  end;
end;

{----------------------------------------------------------------------------
    #1c. TSeries - drawing routines
 ----------------------------------------------------------------------------}

procedure TSeries.RequestPaint;
begin
 if FHoldUpdates
  then updateswaiting := true
  else
   begin
   inherited requestpaint(TheMessage);
   if (TheMessage <> dsUnLoad) and FGraph.fplotting then FGraph.paint;
   end;
end;


type
  pLinePoints = ^TLinePoints;
  TLinePoints = array[0..15000] of TPoint;

procedure TSeries.DrawMyLine;
var
  pp: pgPoint;
  i: Word;
  lp: pLinePoints;
  Zero: double;
  wyax : boolean;
begin
  wyax := (FWhichYAxis = FGraph.FYAxis);
  if FData <> nil then
  with FGraph do
  begin
    FCanvas.Pen.Color := FLineColor;
    FCanvas.Pen.Style := FLineStyle;
    New(lp);
    try
      pp := FData;
      zero := FZeroOffset;
      while pp <> nil do
        begin
        i := 0;
        while (pp <> nil) and (i < 15000) do
             { I routinely turn the i < 15,000 check off for speed. }
          begin
{Jeff Roberts possible bug fix - see the comments in the help under bugs:
          if ((pp^.xv >= fxaxis.fmin) and (pp^.xv <= FXaxis.fmax)) then
          begin}
          lp^[i].x := fx(pp^.xv);
          if wyax
            then lp^[i].y := fy(pp^.yv - Zero)
            else lp^[i].y := fy2(pp^.yv - Zero);
          Inc(i);
          {end;  paired from the begin in brackets above}
          pp := pp^.next;
          end;
        Polyline(FCanvas.Handle, lp^, i);
        end;
    finally
      Dispose(lp);
    end;
  end;
end {DrawMyLine};


procedure TSeries.DrawMyPoints;
var
  s1Size: Word;

  procedure PlotSquare(x, y: Integer);
  begin
    FGraph.FCanvas.Polygon([Pt(x - s1Size, y - s1Size), Pt(x + s1size, y - s1Size),
                             Pt(x + s1Size, y + s1size), Pt(x - s1Size, y + s1Size)]);
  end;

  procedure PlotCircle(x, y: Integer);
  begin
    FGraph.FCanvas.Ellipse(x - s1Size, y - s1Size, x + s1size, y + s1size);
  end;

  procedure PlotDiamond(x, y: Integer);
  begin
    FGraph.FCanvas.Polygon([Pt(x, y - s1Size), Pt(x + s1size, y),
                             Pt(x, y + s1size), Pt(x - s1Size, y)]);
  end;

  procedure PlotCross(x, y: Integer);
  begin
    FGraph.DrawLineSegment(x - s1Size, y, x + s1size, y);
    FGraph.DrawLineSegment(x, y - s1Size, x, y + s1size);
  end;

var
  pp: pgPoint;
  zero: Double;
  ty, OldPenWidth:integer;
  wyax:boolean;
begin {DrawMyPoints}
  wyax := (FWhichYAxis = FGraph.FYAxis);
  if FData <> nil then
  with FGraph do
  begin
    FCanvas.Pen.Color := FPointColor;
    FCanvas.Pen.Style := psSolid;
    FCanvas.Brush.Style := bsSolid;
    if FFillPoints then
      FCanvas.Brush.Color := FPointColor
    else
      FCanvas.Brush.Color := Color;
    if FPointShape = ps_Wide then
    begin
      OldPenWidth := FCanvas.pen.width;
      FCanvas.Pen.Width := Round(FScale*(FPointSize div 2));
      FCanvas.Pen.Color := FPointColor;
    end else OldPenWidth := 0;
    s1Size := Round(FScale*(FPointSize div 2));
    pp := FData;
    zero := FZeroOffset;
    if FAppearance.FAllowedOut then
    begin
      case FPointShape of           { only test this once }
        ps_Square: while pp <> nil do
                   begin
                     if wyax then
                       PlotSquare(fx(pp^.xv), fy(pp^.yv - zero))
                     else
                       PlotSquare(fx(pp^.xv), fy2(pp^.yv - zero));
                     pp := pp^.next;
                   end;
        ps_Circle: while pp <> nil do
                   begin
                     if wyax then
                       PlotCircle(fx(pp^.xv), fy(pp^.yv - zero))
                     else
                       PlotCircle(fx(pp^.xv), fy2(pp^.yv - zero));
                     pp := pp^.next;
                   end;
        ps_Diamond: while pp <> nil do
                   begin
                     if wyax then
                       PlotDiamond(fx(pp^.xv), fy(pp^.yv - zero))
                     else
                       PlotDiamond(fx(pp^.xv), fy2(pp^.yv - zero));
                      pp := pp^.next;
                    end;
        ps_Cross: while pp <> nil do
                  begin
                    if wyax then
                      PlotCross(fx(pp^.xv), fy(pp^.yv - zero))
                    else
                      PlotCross(fx(pp^.xv), fy2(pp^.yv - zero));
                    pp := pp^.next;
                  end;
        ps_Wide: begin
                 if pp = nil then exit;
                 while pp^.next <> nil do
                 begin
                   if wyax then ty := fy(pp^.yv) else ty := fy2(pp^.yv);
                   DrawLineSegment(fx(pp^.xv), ty, fx(pp^.next^.xv), ty);
                   pp := pp^.next;
                 end;
                 if pp^.xv < FXaxis.Max then
                 begin
                   if wyax then ty := fy(pp^.yv) else ty := fy2(pp^.yv);
                   DrawLineSegment(fx(pp^.xv), ty, fx(FXaxis.max), ty);
                 end;
                 end;
      end;
    end
    else { not allowed out }
{   bug fix: 31/5/1997 added -zero at all points where check that the y point is in scale}
    with FGraph do
    begin
      while (pp <> nil) and (pp^.xv < FXAxis.Min) do pp := pp^.next;
        case FPointShape of  { only test this once }
        ps_Square: while (pp <> nil) and (pp^.xv <= FXAxis.Max) do
                   begin
                     if (pp^.yv - zero <= WhichYAxis.Max) and (pp^.yv -zero >= WhichYAxis.Min) then
                       if wyax then
                         PlotSquare(fx(pp^.xv), fy(pp^.yv - zero))
                       else
                         PlotSquare(fx(pp^.xv), fy2(pp^.yv - zero));
                     pp := pp^.next;
                   end;
        ps_Circle: while (pp <> nil) and (pp^.xv <= FXAxis.Max) do
                   begin
                     if (pp^.yv - zero  <= WhichYAxis.Max) and (pp^.yv - zero  >= WhichYAxis.Min) then
                       if wyax then
                         PlotCircle(fx(pp^.xv), fy(pp^.yv - zero))
                       else
                         PlotCircle(fx(pp^.xv), fy2(pp^.yv - zero));
                     pp := pp^.next;
                   end;
        ps_Diamond: while (pp <> nil) and (pp^.xv <= FXAxis.Max) do
                    begin
                      if (pp^.yv - zero <= WhichYAxis.Max) and (pp^.yv - zero >= WhichYAxis.Min) then
                        if wyax then
                          PlotDiamond(fx(pp^.xv), fy(pp^.yv - zero))
                        else
                          PlotDiamond(fx(pp^.xv), fy2(pp^.yv - zero));
                      pp := pp^.next;
                    end;
        ps_Cross: while (pp <> nil) and (pp^.xv <= FXAxis.Max) do
                  begin
                    if (pp^.yv - zero <= WhichYAxis.Max) and (pp^.yv - zero >= WhichYAxis.Min) then
                      if wyax then
                        PlotCross(fx(pp^.xv), fy(pp^.yv - zero))
                      else
                        PlotCross(fx(pp^.xv), fy2(pp^.yv - zero));
                    pp := pp^.next;
                  end;
        ps_Wide: begin
                 if pp = nil then exit;
                 while (pp^.next <> nil) and (pp^.next^.xv - zero < FXAxis.max) do
                 begin
                   if wyax then ty := fy(pp^.yv) else ty := fy2(pp^.yv);
                   DrawLineSegment(fx(pp^.xv), ty, fx(pp^.next^.xv), ty);
                   pp := pp^.next;
                 end;
                 if pp^.xv < FXaxis.Max then
                 begin
                   if wyax then ty := fy(pp^.yv) else ty := fy2(pp^.yv);
                   DrawLineSegment(fx(pp^.xv), ty, fx(FXaxis.max), ty);
                 end;
                 end;
      end;
    end;
    FCanvas.Brush.Color := Color;
    FCanvas.Brush.Style := bsClear;
    if FPointShape = ps_Wide then FCanvas.Pen.Width := OldPenWidth;
  end;
end {DrawMyPoints};


{$IFDEF STATISTICS}
procedure TSeries.DrawMyRegression;
var pp: pgPoint;
    Zero: Double;
    i: Word;
    lp: pLinePoints;
    wyax : boolean;
begin
wyax := (FWhichYAxis = FGraph.FYAxis);
if (FData <> nil) and (FData^.next <> nil) then { no regressions unless at least 2 data points }
 with FGraph do
 begin
   if NeedRegreCalc then DoRegression;
   { can't store smoothing curve (spline, dwls) result with the point - this
     defeats the whole purpose of the spline. So they have to be drawn
     differently to other regression types. for the same reasons, residuals
     are not available like they are with the other regressions - in this
     case the figure returned in getpoint etc. is the second differential
     with spline or nothing with dwls
     The data is available through the RegressionData property if you want it }
   if (FRegType = rg_Spline) or (FRegType = rg_dwls)
     then pp := FRegrData
     else pp := FData;
   FCanvas.Pen.Color := FRegrColor;
   FCanvas.Pen.Style := FRegrLineStyle;
   New(lp);
   try
     Zero := FZeroOffset;
     while (pp <> nil) do
       begin
       i := 0;
       while (pp <> nil) and (i < 15000) do
         {I routinely have the i < 15,000 check turned off}
         begin
         lp^[i].x := fx(pp^.xv);
         if wyax
           then lp^[i].y := fy(pp^.rv - zero)
           else lp^[i].y := fy2(pp^.rv - zero);
         pp := pp^.next;
         Inc(i);
         end;
       Polyline(FCanvas.Handle, lp^, i);
       end;
   finally
     Dispose(lp);
   end;
 end;
end {DrawMyRegression};
{$ENDIF}{STATISTICS}

procedure TSeries.DrawBounds;
  procedure DrawBound(y:double);
  begin
  with FGraph do
    begin
    FCanvas.pen.color := FBoundsColor;
    FCanvas.pen.style := FBoundsLineStyle;
    DrawLineSegment(fx(XMinVal), fy(y), fx(XMaxVal), fy(y));
    end;
  end;
var mean,SD:double;
begin
 if FBoundsType = bt_None then exit;
 if FBoundsType = bt_AsSet then
  begin
  DrawBound(FUpperBound);
  DrawBound(FLowerBound);
  exit;
  end;
 {$IFDEF STATISTICS}
 mean := YStats.Mean;
 SD := YStats.SD; {fix!}
 case FBoundsType of
   bt_1SD  :begin
            drawbound(mean + SD);
            drawbound(mean - SD);
            end;
   bt_2SD  :begin
            drawbound(mean + 2 * SD);
            drawbound(mean - 2 * SD);
            end;
   bt_3SD  :begin
            drawbound(mean + 3 * SD);
            drawbound(mean - 3 * SD);
            end;
   bt_allSD:begin
            drawbound(mean + SD);
            drawbound(mean - SD);
            drawbound(mean + 2 * SD);
            drawbound(mean - 2 * SD);
            drawbound(mean + 3 * SD);
            drawbound(mean - 3 * SD);
            end;
  end;
 {$ENDIF}
end;

procedure TSeries.Draw;
begin
  if not FActive or (FData = nil) then Exit;
  with FGraph do
  begin
    if FXAxis.LogScale or WhichYAxis.LogScale then
    begin
      if self.Changed then GetMinMax;
      if (FXAxis.LogScale and (xvMin <= 0)) or
         (WhichYAxis.LogScale and (yvMin <= 0)) then
      begin
        if Assigned(FOnLogError) then FOnLogError(self, FSeriesIndex);
        Exit;
      end;
    end;
    if not FAppearance.FAllowedOut then ClipGraph;
    DrawBounds;
   {$IFDEF STATISTICS}
    if (FRegType <> rg_None)
      {and possibly allow 1 of these 2: }
      { and not WhichYAxis.LogScale }
      { and not (WhichYAxis.LogScale and
         ((FRegType = rg_Spline) or (FRegType = rg_dwls) or (FRegType = rg_polint)
                or (FRegType = rg_quadratic)) )}
      then DrawMyRegression;

    { Q. What's all this about??
      A. Any of the regression types may cause an exception with ylogging
         turned on. You have 3 options:
        1. Try to draw them anyway, accepting that an exception may occur.
             For this option leave both checks commented out
        2. Be very careful and not allow any to be drawn.
             For this option, allow the second check
        3. Only allow the linear types to be drawn - these are less likely to
           cause an exception.
             For this option, allow the third check but not the second

     !! Note that if a series contains negative data while Ylogging is true,
        the paint will never make it this far   }
   {$ENDIF}
    if FDrawLine then DrawMyLine;
    if not FAppearance.FAllowedOut then UnclipGraph;
   { it'd be nice if we were able to clip the points, but the clipping
      applies to the centre of the point, not the plotted shape }
    if FDrawPoints then DrawMyPoints;
  end;
end {Draw};

{--------------------------------------------------------------------
    #2a. TAxis  - scaling routines
 --------------------------------------------------------------------}

procedure TAxis.InitLogTicks;
begin
  AxisInitLogTicks(FLogTickInfo);
end;

procedure TAxis.AdjustLogTickCount;
{CAREFUL! these LogTickCount values must correspond with AllowedLogTickCounts}
var
  r: Double;
begin
  r := FGraph.FAppearance.MinSteps/(ln(FMax/FMin)*Loge);
  with FLogTickInfo do
    if      r >    10 then LogTickCount := 9
    else if r >     8 then LogTickCount := 7
    else if r >     6 then LogTickCount := 5
    else if r >     3 then LogTickCount := 2
    else if r >     2 then LogTickCount := 1
    else                   LogTickCount := 0;
  InitLogTicks;
end {AdjustLogTickCount};

procedure TAxis.CheckDrawMinorLabels;
begin
  with FLogTickInfo do
  if LogStepping then
    DrawMinorLabels := true
  else
    DrawMinorLabels := ( (ln(FMax/FMin)*Loge)*LogTickCount < 10)
end;

{ step size chosen in a 1,2,5,10 squence depending not only on the
  characteristic, but also the mantissa, of the range}
function TAxis.GetStep: Double;
var
  w, t, B: Double; minSteps: Word;
begin
  minSteps := FGraph.FAppearance.MinSteps;
  w := FMax - FMin;
if w <= 0 then raise Exception.Create('GetStep entered with bad range');
  t := ln(w)*Loge;
  if t < 0 then t := t - 1;
  B := exp( trunc(t * 1.001) / Loge );

  if         w/B >= minSteps then Result := B
  else if  2*w/B >= minSteps then Result := B/2
  else if  5*w/B >= minSteps then Result := B/5
  else if 10*w/B >= minSteps then Result := B/10
  else if 20*w/B >= minSteps then Result := B/20
  else if 50*w/B >= minSteps then Result := B/50
  else                            Result := B/100;
  {sufficient for maxSteps <= 125}
end {GetStep};

function TAxis.getDateStep: double;
begin
{ these are approximations - they are not used for actually
  plotting the ticks, but for checking the scale }
  case FDateTickType of
    dt_minute    :result := cminute1;
    dt_5minute   :result := cminute5;
    dt_halfhourly:result := cminute30;
    dt_hourly    :result := chour;
    dt_daily     :result := 1;
    dt_weekly    :result := 7;
    dt_monthly   :result := 30;
    dt_2monthly  :result := 61;
    dt_quarterly :result := 91;
    dt_annually  :result := 365;
    dt_decade    :result := 3654;
    dt_century   :result := 36540;
    dt_custom    :result := getStep;
  else result := 1;
  end;
end;

{--------- DoResize helper functions -------}
procedure TAxis.SetDateMinMax;
var nYear, nMonth, nDay, nHour, nMin, nSec, nMSec,
    xYear, xMonth, xDay, xHour, xMin, xSec, xMSec: Word;
begin
  if (oMax - oMin <= 0) or (oMin < 1) then exit;
  DecodeDate(oMin, nYear, nMonth, nDay);
  DecodeDate(oMax, xYear, xMonth, xDay);
  DecodeTime(oMin, nHour, nMin, nSec, nMSec);
  DecodeTime(oMax, xHour, xMin, xSec, xMSec);
  case FDateTickType of
   dt_custom:{ in this case no action is taken to change the scale of the axes }
      setMinMax;
   dt_minute:      {set begin to nearest whole minute before start and set end to}
                   {nearest whole minute after end}
      begin
      FMin := EncodeDate(nYear, nMonth, nDay) + EncodeTime(nHour, nMin, 0, 0);
      If xSec <> 0
       then FMax := EncodeDate(xYear, xMonth, xDay)
                      + EncodeTime(xHour, xMin, 0, 0) + encodetime(0,1,0,0)
       else FMax := EncodeDate(xYear, xMonth, xDay)
                      + EncodeTime(xHour, xMin, 0, 0);
      end;
   dt_5minute:begin
      FMin := EncodeDate(nYear, nMonth, nDay) + EncodeTime(nHour, 5 * (nMin div 5), 0, 0);
      If (xMin mod 5 <> 0) or (xSec <> 0)
        then FMax := EncodeDate(xYear, xMonth, xDay) + EncodeTime(xHour,
                        5 * (xMin div 5), 0, 0) + encodetime(0,5,0,0)
        else FMax := EncodeDate(xYear, xMonth, xDay) + EncodeTime(xHour, xMin, 0, 0);
      end;
   dt_halfhourly:begin
      if nMin >= 30
       then FMin := EncodeDate(nYear, nMonth, nDay) + EncodeTime(nHour, 30, 0, 0)
       else FMin := EncodeDate(nYear, nMonth, nDay) + EncodeTime(nHour, 0, 0, 0);
      If (xMin = 0) and (xSec = 0)
       then FMax := EncodeDate(xYear, xMonth, xDay) + EncodeTime(xHour, 0, 0, 0)
       else if (xMin >= 30)
        then FMax := EncodeDate(xYear, xMonth, xDay) + EncodeTime(xHour, 0, 0, 0)
                      + encodetime(1,0,0,0)
        else FMax := EncodeDate(xYear, xMonth, xDay) + EncodeTime(xHour, 30, 0, 0);
      end;
   dt_hourly:begin
      Fmin := EncodeDate(nYear, nMonth, nDay) + EncodeTime(nHour, 0, 0, 0);
      if (xSec <> 0) or (xMin <> 0)
       then FMax := EncodeDate(xYear, xMonth, xDay) + EncodeTime(xHour, 0, 0, 0)
                  + encodetime(1,0,0,0)
       else FMax := EncodeDate(xYear, xMonth, xDay) + EncodeTime(xHour, 0, 0, 0);
      end;
   dt_daily:begin
      FMin :=  EncodeDate(nYear, nMonth, nDay);
      if (xHour <> 0) or (xMin <>0) or (xSec <> 0)
       then FMax := EncodeDate(xYear, xMonth, xDay) + 1
       else FMax := EncodeDate(xYear, xMonth, xDay);
      end;
   dt_weekly:begin
      FMin := EncodeDate(nYear, nMonth, nDay) - (DayOfWeek(oMin) - 1);
      FMax := EncodeDate(xYear, xMonth, xDay) + 8 - DayOfWeek(oMin); {???}
      end;
   dt_monthly:begin
      FMin :=  EncodeDate(nYear, nMonth, 1);
      if (xday > 1) or (xHour <> 0) or (xMin <>0) or (xSec <> 0)
       then incmonth(xYear,xMonth,1); {month is not set time period}
      FMax := EnCodeDate(xYear,xMonth, 1);
      end;
   dt_2monthly:begin
      if nMonth mod 2 <> 1 then dec(nMonth);
      FMin :=  EncodeDate(nYear, nMonth, 1);
      if (xday > 1) or (xHour <> 0) or (xMin <>0) or (xSec <> 0)
       then incmonth(xYear,xMonth,1); {month is not set time period}
      if xMonth mod 2 <> 1 then inc(xMonth);
      FMax := EnCodeDate(xYear,xMonth, 1);
      end;
   dt_quarterly:begin
      while nMonth mod 3 <> 1 do dec(nMonth);
      FMin :=  EncodeDate(nYear, nMonth, 1);
      if (xday > 1) or (xHour <> 0) or (xMin <>0) or (xSec <> 0)
        then incmonth(xYear,xMonth,1); {month is not set time period}
      while xMonth mod 3 <> 1 do incmonth(xYear,xMonth,1);
      FMax := EnCodeDate(xYear,xMonth, 1);
      end;
   dt_annually:begin
      FMin := EncodeDate(nYear, 1, 1);
      if (xMonth > 1) or(xday > 1) or (xHour <> 0) or (xMin <>0) or (xSec <> 0)
       then FMax := EncodeDate(xYear+1,1,1)
                  { year is also not a set time,  but there can't be overflow here }
       else FMax := EnCodeDate(xYear,1,1);
      end;
   dt_decade:begin
      FMin := EncodeDate(10 * (nYear div 10), 1, 1);
      if (xMonth > 1) or(xday > 1) or (xHour <> 0) or (xMin <>0) or (xSec <> 0)
        then inc(xYear);
      FMax := EncodeDate(10 * (xYear div 10),1,1);
      end;
   dt_century:begin
      FMin := EncodeDate(100 * (nYear div 100), 1, 1);
      if (xMonth > 1) or(xday > 1) or (xHour <> 0) or (xMin <>0) or (xSec <> 0)
        then inc(xYear);
      FMax := EncodeDate(100 * (xYear div 100),1,1);
      end;
  end;
end;

procedure TAxis.SetMinMax;
begin
if oMin > oMax then raise Exception.Create('Impossible: oMin > oMax in SetMinMax!');
  if (oMin = oMax) then
  begin
    if (oMin <> 0) then
    begin
      if (oMin < 0) then
      begin
        FMin := oMin * 1.2;
        FMax := 0;
      end
      else
      begin
        FMin := 0;
        FMax := oMax * 1.2;
      end;
      FStep := GetStep;
      FMax := trunc(FMax / FStep) * FStep;
    end
    else
    begin
      FMin := -1;
      FMax := 1;
      FStep := GetStep;
    end;
    Exit;
  end
  else
  begin
    FMin := oMin;
    FMax := oMax;
  end;
  {assert: FMin<FMax}
  if FAutoStepping or (FStep <= 0) or ( FStep/(FMax-FMin) < 1/FGraph.appearance.FMaxSteps ) then
    FStep := GetStep;

  FMin := trunc( FMin / FStep) * FStep;
  if (oMin < 0) then FMin := FMin - FStep;
  if (FMin > oMin - 0.01*FStep) then FMin := FMin - FStep;

  FMax := trunc(FMax / FStep) * FStep + FStep;
  if (oMax < 0) then FMax := FMax - FStep;
  if (FMax < oMax + 0.01*FStep) then FMax := FMax + FStep;

  if (oMin > 0) and (oMin /(FMax - FMin) < 0.25) then
    FMin := 0
  else if (oMax < 0) and (-oMax / (FMax - FMin) < 0.25) then
    FMax := 0;
if Fmin >= FMax then raise Exception.Create('SetMinMax failed');
end {SetMinMax};


procedure TAxis.SetLogMinMax;
const
  above = true;
  below = not above;

  function GetLogTick(b: Boolean; v: Double): Double;
  var
    ix: Word;
  begin
    with FLogTickInfo do
    begin
      if b = above then
      begin
        ix := 1;
        while (ix < LogTickCount+1) and
              (LogTickArray[ix] < v) do Inc(ix);
      end
      else {below}
      begin
        ix := LogTickCount;
        while (ix > 0) and
              (LogTickArray[ix] > v) do Dec(ix);
      end;
      Result := LogTickArray[ix];
    end;
  end {GetLogTick};

var
  minChi, maxChi, minB, maxB, t: Double;
  minN, maxN : Integer;

begin {SetLogMinMax}
  if (oMin = oMax) then
  begin
    FMin := 0.5*oMax;
    FMax := oMax;
  end
  else
  begin
    FMin := oMin;
    FMax := oMax;
  end;

  t := ln(FMin) * Loge;
  minN := trunc(t);
  if (t < 0) and (t <> minN) then Dec(minN);
  minChi := exp( (t - minN) / Loge );
  minB := exp(minN/Loge);

  t := ln(FMax) * Loge;
  maxN := trunc(t);
  if ((t < 0) and (t <> maxN)) or (t = 0) then Dec(maxN);
  maxChi := exp( (t - maxN) / Loge );
  maxB := exp(maxN/Loge);

  if FAutoStepping then
    AdjustLogTickCount;

  if (FMax/FMin < 4) then
  begin
    {need to set up for constant ticks across FMin..FMax}
    FLogTickInfo.LogStepping := true;
    SetMinMax;
  end
  else
  begin
    FLogTickInfo.LogStepping := false;
    if (FMax/FMin < 10) then
    begin
      FMin := minB*GetLogTick(below, minChi);
      FMax := maxB*GetLogTick(above, maxChi);
    end
    else  if (FMax/FMin < 30) then
    begin
      FMin := GetLogTick(below, minChi);
      if FMin > 1 then FMin := GetLogTick(below, FMin);
      FMin := minB*FMin;
      FMax := GetLogTick(above, maxChi);
      if FMax < 10 then FMax := GetLogTick(above, FMax);
      FMax := maxB*FMax;
    end
    else
    begin
      FMin := minB;
      FMax := maxB*10;
    end;
  end;
  CheckDrawMinorLabels;
end {SetLogMinMax};


function TAxis.DoResize: Boolean;

  function ArbitraryMinMax: Boolean;
  var hold: Double;
  begin
    Result := true;
    if not FLogging then
      if FShowAsTime then
      begin
      FMax := now;
      Fmin := FMax - 5 * GetDateStep;
      end
      else
      begin
        if (FMin = FMax) then
        begin
          if (FMin <> 0) then
            FMin := 0
          else
          begin
            FMin := -1;
            FMax := 1;
          end;
        end
        else if (FMin > FMax) then
        begin
          hold := FMin;
          FMin := FMax;
          FMax := hold;
        end
        else
          Result := false;
        FStep := GetStep;
    end
    else {logging}
    begin
      Result := false;
      if (FMin > FMax) then
      begin
        hold := FMin;
        FMin := FMax;
        FMax := hold;
        Result := true;
      end;
      if (FMax <= 0) then
      begin
        FMax := 10;
        FMin := 1;
      end
      else if (FMin <= 0) then
        FMin := FMax/10
      else
        Result := false or Result;
    end
  end {ArbitraryMinMax};

begin {DoResize}
  if oMax - oMin < FMinScale then oMax := oMin + FMinScale;
  Result := false;
  if not FGraph.DataFound(self) then
  begin
    Result := ArbitraryMinMax;
    if FLogging then
      CheckDrawMinorLabels;
    Exit;
  end {not datafound};

 {data: pick intelligent min/max/tick values:}
  if FAutoSizing then
  begin
    if not FLogging then
      if FShowAsTime then
        SetDateMinMax
      else
      begin
        SetMinMax;
        if FAutoStepping and (oMin < oMax) then
        begin
          FStep := GetStep;
          SetMinMax;
        end;
      end
    else
    begin {logging}
      if (oMin <= 0) then Exit; {nothing can be done}
      SetLogMinMax;
      if FAutoStepping and (oMin < oMax) and (not FLogTickInfo.LogStepping) then
      begin
        AdjustLogTickCount;
        SetLogMinMax;
      end;
    end;
    Result := (FMin <> oMin) or (FMax <> oMax);
  end;
end {DoResize};

function OneSigFigDecs(v: Double): Integer;
{eg: OneSigFigDecs(0.1)->1; (100)-> -2}
var
  t: Double;
begin
  t := ln(v*1.01)*Loge;
  if t < 0 then
    Result  := -trunc(t) + 1
  else
    Result  := -trunc(t);
end;


procedure  TAxis.AdjustLabelDecs;
begin
  if Not FAutoLabelDecimals then exit;
  FLabelDec := OneSigFigDecs(FStep*1.01);
  if FLabelDec < 0 then
    FLabelDec := 0;
end;

procedure  TAxis.AdjustLogLabelDecs(v: Double);
{the >= 5 must correspond with InitLogTicks: where do fraction steps start?}
begin
  FLabelDec := OneSigFigDecs(v*1.01);
  if FLogTickInfo.LogTickCount >= 5 then Inc(FLabelDec);
  if FLabelDec < 0 then
    FLabelDec := 0;
end;

function TAxis.GetFirstDateTick:double;
var Year, Month, Day, Hour, mMin, Sec, MSec : Word;
begin
  DecodeDate(Fmin, Year, Month, Day);
  DecodeTime(FMin, Hour, mMin, Sec, MSec);
  case FDateTickType of
   dt_custom:result := FMin;
   dt_minute:If Sec <> 0 then
              result := EncodeDate(Year, Month, Day) + EncodeTime(Hour, mMin, 0, 0)
                  + encodetime(0,1,0,0) else
              result := EncodeDate(Year, Month, Day) + EncodeTime(Hour, mMin, 0, 0);
   dt_5minute:If (mMin mod 5 <> 0) or (Sec <> 0) then
              result := EncodeDate(Year, Month, Day) + EncodeTime(Hour,
                        5 * (mMin div 5), 0, 0) + encodetime(0,5,0,0)  else
              result := EncodeDate(Year, Month, Day) + EncodeTime(Hour, mMin, 0, 0);
   dt_halfhourly:If (mMin = 0) and (Sec = 0) then
              result := EncodeDate(Year, Month, Day) + EncodeTime(Hour, 0, 0, 0)
                 else if (mMin >= 30) then
              result := EncodeDate(Year, Month, Day) + EncodeTime(Hour, 0, 0, 0)
               + encodetime(1,0,0,0) else
              result := EncodeDate(Year, Month, Day) + EncodeTime(Hour, 30, 0, 0);
   dt_hourly:if (Sec <> 0) or (mMin <> 0) then
              result := EncodeDate(Year, Month, Day) + EncodeTime(Hour, 0, 0, 0)
                  + encodetime(1,0,0,0) else
              result := EncodeDate(Year, Month, Day) + EncodeTime(Hour, 0, 0, 0);
   dt_daily:if (Hour <> 0) or (mMin <>0) or (Sec <> 0) then
              result := EncodeDate(Year, Month, Day) + 1 else
              result := EncodeDate(Year, Month, Day);
   dt_weekly:result := EncodeDate(Year, Month, Day) + 8 - DayOfWeek(oMin); {???}
   dt_monthly:begin
             if (Day > 1) or (Hour <> 0) or (mMin <>0) or (Sec <> 0) then
                incmonth(Year,Month,1); {month is not set time period}
             result := EnCodeDate(Year,Month, 1);
             end;
   dt_2monthly:begin
               if (Day > 1) or (Hour <> 0) or (mMin <>0) or (Sec <> 0) then
                 incmonth(Year,Month,1); {month is not set time period}
               if Month mod 2 <> 1 then inc(Month);
               result := EnCodeDate(Year,Month, 1);
               end;
   dt_quarterly:begin
               if (Day > 1) or (Hour <> 0) or (mMin <>0) or (Sec <> 0) then
                 incmonth(Year,Month,1); {month is not set time period}
               while Month mod 3 <> 1 do incmonth(Year,Month,1);
               result := EnCodeDate(Year,Month, 1);
               end;
   dt_annually:begin
               if (Month > 1) or(Day > 1) or (Hour <> 0) or
                     (mMin <>0) or (Sec <> 0) then
                result := EncodeDate(Year+1,1,1)
                  { year is also not a set time,  but there can't be overflow here }
               else result := EnCodeDate(Year,1,1);
               end;
   dt_decade:begin
               if (Month > 1) or(Day > 1) or (Hour <> 0) or
                     (mMin <>0) or (Sec <> 0) then  inc(Year);
               result := EncodeDate(10 * (Year div 10),1,1);
               end;
   dt_century:begin
               if (Month > 1) or(Day > 1) or (Hour <> 0) or
                     (mMin <>0) or (Sec <> 0) then  inc(Year);
               result := EncodeDate(100 * (Year div 100),1,1);
               end;
   else result := 1;
  end;
end;

function TAxis.GetFirstTick(var logTickInfo: TLogTickInfo): Double;
var
  t, B: Double;
  j: Word;
begin
  if not FLogging or logTickInfo.LogStepping then
  if FShowAsTime then result := getfirstdatetick else
  begin
    Result := FMin + 0.01*FStep;
    Result := trunc( Result / FStep ) * FStep;
    if (FMin < 0) then Result := Result - FStep;
    if (FMin > Result + 0.01*FStep) then Result := Result + FStep;
    AdjustLabelDecs;
  end
  else {logging}
  begin
    t := ln(FMin) * Loge;
    if t < 0 then t := t - 1;
    B := exp( trunc(1.001 * t) / Loge ); { OK for FMin < 10^1000 }
    if B > FMin then B := B/10;

    {pre-condition: the TLogTickInfo has been initialized with 1..10:}
    with logTickInfo do
    begin
      for j := 0 to LogTickCount+1 do
        LogTickArray[j] := B * LogTickArray[j];

      LogTickIndex := 0;
      t := FMin*0.999;
      while logTickInfo.LogTickArray[LogTickIndex] < t do
        Inc(LogTickIndex);
      Result := LogTickArray[LogTickIndex];
      AdjustLogLabelDecs(Result);
    end;
  end;
end {GetFirstTick};


function TAxis.GetNextDateTick(tick: Double):double;
var year,month,day:word;
begin
 case FDateTickType of
  dt_minute: result := tick + cminute1;
  dt_5minute: result := tick + cminute5;
  dt_halfhourly: result := tick + cminute30;
  dt_hourly: result := tick + chour;
  dt_daily: result := tick + 1;
  dt_weekly: result := tick + 7;
  dt_monthly:begin
             decodedate(tick,year,month,day);
             incmonth(year,month,1);
             result := encodedate(year, month, 1);
             end;
  dt_2monthly:begin
             decodedate(tick,year,month,day);
             incmonth(year,month,2);
             result := encodedate(year, month, 1);
             end;
  dt_quarterly:begin
             decodedate(tick,year,month,day);
             incmonth(year,month,3);
             result := encodedate(year, month, 1);
             end;
  dt_annually:begin
             decodedate(tick,year,month,day);
             result := encodedate(year+1, 1, 1);
             end;
  dt_decade:begin
             decodedate(tick,year,month,day);
             result := encodedate(year+10, 1, 1);
             end;
  dt_century:begin
             decodedate(tick,year,month,day);
             result := encodedate(year+100, 1, 1);
             end;
  dt_custom:Result := tick + FStep;
  else result := tick+FStep;
 end;
end;

function TAxis.GetNextTick(tick: Double; var logTickInfo: TLogTickInfo; var drawThisLabel: Boolean): Double;
var j: Word;
begin
  if not FLogging or logTickInfo.LogStepping then
    if FShowAsTime then
    begin
      result := getnextdateTick(tick);
      drawthislabel := true;
    end
    else
    begin
      Result := tick + FStep;
      drawThisLabel := true;
    end
  else {logging}
    with logTickInfo do
    begin
      Inc(LogTickIndex);
      if LogTickIndex >= LogTickCount + 1 then
      begin
        for j := 0 to LogTickCount + 1 do
          LogTickArray[j] := 10 * LogTickArray[j];
        LogTickIndex := 0;
      end;
      Result := LogTickArray[LogTickIndex + 1];
      if (LogTickIndex = 0) then
        AdjustLogLabelDecs(Result);
      drawThisLabel := DrawMinorLabels or
                       (LogTickIndex = 0) or
                       (LogTickIndex = LogTickCount+1);
    end {with, else};
end {GetNextTick};


function TAxis.CheckScale: Boolean;
begin
  Result := (FMin < FMax) and ((not FLogging) or (FMin > 0)) and
              not (FShowAsTime and (Fmin < 1));
  scaledOK := result;
end;

procedure TAxis.CalcMetrics;
begin
 if (FMin>=FMax) then raise Exception.Create(FAxisTitle+'CalcMetrics: XMin>=XMax');
 with FGraph do
 if self = FGraph.FXAxis then
  if FLogging then
    FM := (FCurrWidth - FDimensions.FLeft - FDimensions.FRight) / (Ln(FMax) - Ln(FMin))
  else
    FM := (FCurrWidth - FDimensions.FLeft - FDimensions.FRight) / (FMax - FMin)
 else
  if FLogging then
    FM := (FCurrHeight - FDimensions.FTop - FDimensions.FBottom) / (Ln(FMax) - Ln(FMin))
  else
    FM := (FCurrHeight - FDimensions.FTop - FDimensions.FBottom) / (FMax - FMin);
end;


{--------------------------------------------------------------------
    #2b. TAxis - property servers
 --------------------------------------------------------------------}

procedure TAxis.SetAxisTitle(v: string);
begin
  FAxisTitle := v;
  if FGraph.fplotting then FGraph.Paint;
end;

procedure TAxis.SetLabelDec(v: Integer);
begin
  FLabelDec := v;
  if FGraph.fplotting then FGraph.Paint;
end;

procedure TAxis.SetLogTickCount(v: Word);
begin
  FLogTickInfo.LogTickCount := v;
  InitLogTicks;
  if FGraph.fplotting then FGraph.Paint;
end;

procedure TAxis.setdateticktype;
begin
  if v <> FDateTickType then
    begin
    FDateTickType := v;
    if FGraph.fplotting then FGraph.Paint;
    end;
end;

function TAxis.GetLogTickCount: Word;
begin
  Result := FLogTickInfo.LogTickCount;
end;

procedure TAxis.SetLogging(v: Boolean);
begin
  FLogging := v;
  if FGraph.fplotting then FGraph.Paint;
end;

procedure TAxis.SetAutoSizing(v: Boolean);
begin
  FAutoSizing := v;
  if FGraph.fplotting then FGraph.Paint;
end;

procedure TAxis.SetAutoStepping(v: Boolean);
begin
  FAutoStepping := v;
  if FGraph.fplotting then FGraph.Paint;
end;

procedure TAxis.SetMax(v: Double);
begin
  FAutoSizing := False;
  if FMax <> v then
  begin
    FMax := v;
    if FShowASTime and not FGraph.datafound(self) then oMax := FMax;
    FGraph.DoRescaleEvent;
    if FGraph.fplotting then FGraph.Paint;
  end;
end;

procedure TAxis.SetMin(v: Double);
begin
  FAutoSizing := False;
  if FMin <> v then
  begin
    FMin := v;
    if FShowASTime and not FGraph.datafound(self) then oMin := FMin;
    FGraph.DoRescaleEvent;
    if FGraph.fplotting then FGraph.Paint;
  end;
end;

procedure TAxis.SetStep(v: Double);
begin
  FAutoStepping := False;
  if FStep <> v then
  begin
    FStep := v;
    FGraph.DoRescaleEvent;
    if FGraph.fplotting then FGraph.Paint;
  end;
end;

procedure TAxis.SetMinDiff(v: Double);
var holdmin: Double;
begin
  holdmin := FMinScale;
  FMinScale := v;
  if FGraph.Plotting and ((FMax - FMin < FMinScale) or (FMax - FMin = holdmin))
  then FGraph.Paint;
end;

procedure TAxis.setShowASTime;
begin
  if v <> FShowASTime then
  begin
    FShowAsTime := v;
    if FShowASTime and not FGraph.datafound(self) then
      begin
      oMin := FMin;
      oMax := FMax;
      end;
    if FGraph.Plotting then FGraph.paint;
  end;
end;

procedure TAxis.SetDateFormat;
begin
  if v <> FDateFormat then
    begin
    FDateFormat := v;
    if FGraph.Plotting then FGraph.Paint;
    end;
end;

procedure TAxis.SetShowAxis;
begin
  if v <> FShowAxis then
    begin
    FShowAxis := v;
    if FGraph.Plotting then FGraph.Paint;
    end;
end;

procedure TAxis.SetReversed;
begin
  if v <> FReversed then
    begin
    FReversed := v;
    if FGraph.Plotting then FGraph.Paint;
    end;
end;

procedure TAxis.setOffsetType;
begin
  if v <> FOffsetType then
    begin
    FOffsetType := v;
    if FGraph.Plotting then FGraph.Paint;
    end;
end;

procedure TAxis.setoffset;
begin
  if v <> FOffset then
    begin
    FOffset := v;
    if FGraph.Plotting then FGraph.Paint;
    end;
end;

procedure TAxis.SetGridlines(v: Boolean);
begin
  FGridlines := v;
  if FGraph.fplotting then FGraph.Paint;
end;

{-------------------------------------------------------------------
    #3. TDimensions
--------------------------------------------------------------------}

procedure TDimensions.SetMargBottom(v: Word);
begin
  FBottom := v;
  if FGraph.fplotting then FGraph.Paint;
end;

procedure TDimensions.SetMargTop(v: Word);
begin
  FTop := v;
  if FGraph.fplotting then FGraph.Paint;
end;

procedure TDimensions.SetMargLeft(v: Word);
begin
  FLeft := v;
  if FGraph.fplotting then FGraph.Paint;
end;

procedure TDimensions.SetMargRight(v: Word);
begin
  FRight := v;
  if FGraph.fplotting then FGraph.Paint;
end;

procedure TDimensions.SetTickLength(v: Word);
begin
  FTMLength := v;
  if v = 0 then FGraph.FAppearance.FTickMarks := False;
  if FGraph.fplotting then FGraph.Paint;
end;

procedure TDimensions.SetGraphTitleDistance(v: Integer);
begin
  FGraphTitleDistance := v;
  if FGraph.fplotting then FGraph.Paint;
end;

procedure TDimensions.SetXAxisTitleDistance(v: Integer);
begin
  FXAxisTitleDistance := v;
  if FGraph.fplotting then FGraph.Paint;
end;

procedure TDimensions.SetYAxisTitleDistance(v: Integer);
begin
  FYAxisTitleDistance := v;
  if FGraph.fplotting then FGraph.Paint;
end;

procedure TDimensions.SetXAxisLabelDistance(v: Integer);
begin
  FXAxisLabelDistance := v;
  if FGraph.fplotting then FGraph.Paint;
end;

procedure TDimensions.SetYAxisLabelDistance(v: Integer);
begin
  FYAxisLabelDistance := v;
  if FGraph.fplotting then FGraph.Paint;
end;

procedure TDimensions.SetScale(v: Word);
begin
  if v < 20 then v := 20 else if v > 100 then v := 100;
  FScalePct := v;
end;

procedure TDimensions.SetXOffset(v: Word);
begin
  if v < 0 then v := 0 else if v > 100 then v := 100;
  FXOffsetPct := v;
end;

procedure TDimensions.SetYOffset(v: Word);
begin
  if v < 0 then v := 0 else if v > 100 then v := 100;
  FYOffsetPct := v;
end;

{-------------------------------------------------------------------
    #4. TAppearance
---------------------------------------------------------------------}
procedure TAppearance.SetAllowDuplicates;
var p:TSeries;
begin
  FAllowDuplicates := v;
  p := Fgraph.FSeries;
  while p <> nil do
  begin
    p.Allowduplicates := FAllowDuplicates;
    p := p.next;
  end;
  Fgraph.Paint;
end;

procedure TAppearance.SetAllowedOut;
begin
  FAllowedOut := v;
  if FGraph.fplotting then FGraph.Paint;
end;

procedure TAppearance.SetLabelGraph(v: Boolean);
begin
  FLabelGraph := v;
  if FGraph.fplotting then FGraph.Paint;
end;

procedure TAppearance.SetTickmarks(v: Boolean);
begin
  FTickMarks := v;
  if FGraph.fplotting then FGraph.Paint;
end;

procedure TAppearance.SetShowMarks;
begin
  if v <> FShowMarks then
  begin
    FShowMarks := v;
    if FGraph.fplotting then FGraph.Paint;
  end;
end;

procedure TAppearance.SetGridcolor(v: TColor);
begin
  FGridColor := v;
  if FGraph.fplotting then FGraph.Paint;
end;

procedure TAppearance.SetAxescolor(v: TColor);
begin
  FAxesColor := v;
  if FGraph.fplotting then FGraph.Paint;
end;

procedure TAppearance.SetBkgdcolor(v: TColor);
begin
  FBkgdColor := v;
  if FGraph.fplotting then FGraph.Paint;
end;

procedure TAppearance.SetMarginColor(v: TColor);
begin
  FMarginColor := v;
  if FGraph.fplotting then FGraph.Paint;
end;

procedure TAppearance.SetGridStyle(v: TPenStyle);
begin
  FGridStyle := v;
  if FGraph.fplotting then FGraph.Paint;
end;

procedure TAppearance.SetErrorCaption;
begin
  FErrorCaption := v;
  if not FGraph.CheckScales then if FGraph.fplotting then FGraph.Paint;
end;

procedure TAppearance.SetGraphTitle;
begin
  FGraphTitle := v;
  if FGraph.fplotting then FGraph.Paint;
end;

procedure TAppearance.SetMinSteps;
begin
  if (v > 0) and (v < FMaxSteps/2.5) then
    FMinSteps := v;
  if FGraph.fplotting then FGraph.Paint;
end;

procedure TAppearance.SetMaxSteps;
begin
  if (v > FMinSteps*2.5) and (v <= 90) then
    FMaxSteps := v;
  if FGraph.fplotting then FGraph.Paint;
end;

procedure TAppearance.setTitleFont(v:TFont);
begin
  FTitleFont.assign(v);
  if FGraph.fplotting then FGraph.Paint;
end;

procedure TAppearance.setCaptionFont(v:TFont);
begin
  FCaptionFont.assign(v);
  if FGraph.fplotting then FGraph.Paint;
end;

procedure TAppearance.setLabelFont(v:TFont);
begin
  FLabelFont.assign(v);
  if FGraph.fplotting then FGraph.Paint;
end;

procedure TAppearance.setCrossAtZero;
begin
  if v <> FCrossAtZero then
    begin
    FCrossAtZero := v;
    if FGraph.fplotting then FGraph.Paint;
    end;
end;

procedure TAppearance.SetCrossColor;
begin
  if v <> FCrossColor then
    begin
    FCrossColor := v;
    if FGraph.fplotting then FGraph.Paint;
    end;
end;

procedure TAppearance.SetCrossLength;
begin
  if v <> FCrossLength then
    begin
    FCrosslength := v;
    if FGraph.fplotting then FGraph.Paint;
    end;
end;

procedure TAppearance.FontChanged(Sender: TObject);
begin
  if FGraph.Fplotting and (FGraph.FCanvas <> Printer.Canvas) then Fgraph.paint;
end;

{-------------------------------------------------------------------
    #4. TLegend
---------------------------------------------------------------------}

procedure TLegend.setvisible(v:Boolean);
begin
  if v <> FVisible then
    begin
    FVisible := v;
    if FGraph.fplotting then fgraph.paint;
    end;
end;

procedure TLegend.setcolor(v:TColor);
begin
  if v <> FColor then
    begin
    FColor := v;
    if FGraph.fplotting then fgraph.paint;
    end;
end;

procedure TLegend.setborderStyle(v:TBorderStyle);
begin
  if v <> Fborderstyle then
    begin
    FBorderstyle := v;
    if FGraph.fplotting then fgraph.paint;
    end;
end;

procedure TLegend.setLayout(v:TLegendStyle);
begin
  if v <> FLayout then
    begin
    Flayout := v;
    if FGraph.fplotting then fgraph.paint;
    end;
end;

procedure TLegend.settop(v:integer);
begin
  if v <> Ftop then
    begin
    FTop := v;
    if FGraph.fplotting then fgraph.paint;
    end;
end;

procedure TLegend.setleft(v:integer);
begin
  if v <> Fleft then
    begin
    FLeft := v;
    if FGraph.fplotting then fgraph.paint;
    end;
end;

procedure TLegend.setwidth(v:integer);
begin
  if v <> Fwidth then
    begin
    fwidth := v;
    if FGraph.fplotting then fgraph.paint;
    end;
end;

procedure TLegend.setheight(v:integer);
begin
  if v <> Fheight then
    begin
    Fheight := v;
    if FGraph.fplotting then fgraph.paint;
    end;
end;

procedure TLegend.setSymbolSpace(v:integer);
begin
  if v <> FSymbolSpace then
    begin
    FSymbolSpace := v;
    if FGraph.fplotting then fgraph.paint;
    end;
end;

procedure TLegend.setXMargin(v:integer);
begin
  if v <> FXMargin then
    begin
    FXMargin := v;
    if FGraph.fplotting then fgraph.paint;
    end;
end;

procedure TLegend.setYMargin(v:integer);
begin
  if v <> FYMargin then
    begin
    FYMargin := v;
    if FGraph.fplotting then fgraph.paint;
    end;
end;

procedure TLegend.setfont(v:TFont);
begin
 FFont.assign(v);    {procedure below hooks FFont.OnChange}
end;

procedure TLegend.Fontchanged;
begin
 if FGraph.fplotting and (fgraph.FCanvas <> Printer.Canvas) then
   FGraph.paint;
end;

{---------------------------------------------------------------------
    #6a. TxyGraph  - administration
 ---------------------------------------------------------------------}

constructor TxyGraph.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  if (csDesigning in ComponentState) then Caption := name else Caption := '';
  FCanvas := Canvas;
  FAltCanvas := Canvas;
  FSeries := nil;
  FLastSeries := 0;
  FHookedSeries := nil;
  FMarks := nil;
  FScale := 1.0;
  FXOffset := 0;
  FYOffset := 0;
  FDragging := false;
  FLegDragging := false;
  FDraggable := true;
  FNoGridlines := false;
  Width := 200;
  Height := 200;
  BevelInner := bvNone;
  BevelOuter := bvNone;
  BorderStyle := bsSingle;
  Font.Color := clBlack;
  Font.Height := - 9;
  Font.Name := 'Arial';
  Font.Style := [];
  Color := clWhite;

  FXAxis := TAxis.Create;
  FXAxis.FGraph := self;
  FYAxis := TAxis.Create;
  FYAxis.FGraph := self;
  FYAxis2 := TAxis.Create;
  FYAxis2.FGraph := self;
  FDimensions := TDimensions.Create;
  FDimensions.FGraph := self;
  FAppearance := TAppearance.Create;
  FAppearance.FGraph := self;
  with FAppearance do
    begin
    FTitleFont := TFont.Create;
    FTitleFont.assign(font);
    FTitleFont.OnChange := FontChanged;
    FLabelFont := TFont.Create;
    FLabelFont.assign(font);
    FLabelFont.OnChange := FontChanged;
    FCaptionFont := TFont.Create;
    FCaptionFont.assign(font);
    FCaptionFont.OnChange := FontChanged;
    end;
  FLegend := TLegend.create;
  with FLegend do
    begin
    FGraph := self;
    Ffont := TFont.create;
    FFont.assign(self.font);
    Ffont.Onchange := fontchanged;
    end;

  FPlotting := True;
  setDefaultPropertyValues;
  FSaveData := true;
end {Create};

procedure TxyGraph.setDefaultPropertyValues;
begin
  with FXAxis do
  begin
    SecondAxis := false;
    FShowAxis := true;
    Freversed := false;
    FMinScale := defMinXScale;
    FMin := 0;
    FMax := 1.0;
    FStep := FMax / 5;
    FLogTickInfo.LogTickCount := 2;
    InitLogTicks;
    FAutoSizing := False;
    FAutoStepping := False;
    FLogging := False;
    FAxisTitle := 'X axis';
    FGridlines := True;
    FLabelDec := 1;
    FAutoLabelDecimals := true;
    FShowASTime := false;
    FDateFormat := 'dd-mmm';
    FOffsetType := ao_Minimum;
    FOffset := 0;
  end;

  with FYAxis do
  begin
    SecondAxis := false;
    FShowAxis := true;
    FReversed := false;
    FMinScale := defMinYScale;
    FMin := 0;
    FMax := 1.0;
    FStep := FMax / 5;
    FLogTickInfo.LogTickCount := 2;
    InitLogTicks;
    FAutoSizing := False;
    FAutoStepping := False;
    FLogging := False;
    FGridlines := True;
    FAxisTitle := 'Y axis';
    FLabelDec := 1;
    FAutoLabelDecimals := true;
    FShowASTime := false;
    FDateFormat := 'dd-mmm';
    FOffsetType := ao_Minimum;
    FOffset := 0;
  end;

  with FYAxis2 do
  begin
    SecondAxis := true;
    FShowAxis := false;
    FReversed := true;
    FMinScale := defMinYScale;
    FMin := 0;
    FMax := 1.0;
    FStep := FMax / 5;
    FLogTickInfo.LogTickCount := 2;
    InitLogTicks;
    FAutoSizing := False;
    FAutoStepping := False;
    FLogging := False;
    FGridlines := False;
    FAxisTitle := 'Second Y axis';
    FLabelDec := 1;
    FShowASTime := false;
    FDateFormat := 'dd-mmm';
    FOffsetType := ao_Maximum;
    FOffset := 0;
  end;

  with FDimensions do
  begin
    FBottom := 40;
    FLeft := 40;
    FRight := 15;
    FTop := 30;
    FTMLength := 4;
    FGraphTitleDistance := 7;
    FXAxisTitleDistance := 4;
    FXAxisLabelDistance := 2;
    FYAxisTitleDistance := 4;
    FYAxisLabelDistance := 2;
    FScalePct := 90;
    FXOffsetPct := 5;
    FYOffsetPct := 20;
  end;

  with FAppearance do
  begin
    FAllowDuplicates := False;
    FLabelGraph := True;
    FShowMarks := False;
    FTickMarks := True;
    FAllowedOut := False;
    FErrorCaption := 'MAX/MIN VALUES INCONSISTENT';
    FAxesColor := clBlack;
    FGridColor := clSilver;
    FGridStyle := psdot;
    FBkgdColor := color;
    FMarginColor := color;
    FGraphTitle := 'xyGraph ' + versionstamp;
    FMinSteps := 5;
    FMaxSteps := 50;
    FCrossAtZero := true;
    FCrossColor := clred;
    FCrosslength := 4;
    FPrintLineStyle := true;
  end;

  with FLegend do
    begin
    Fvisible := false;
    Fcolor := FAppearance.FBkGdcolor;
    FborderStyle := bsNone;
    FLayout := lsAcross;
    Ftop := 10;
    Fleft := 10;
    Fwidth := self.width - 20;
    Fheight := 20;
    FSymbolSpace := 20;
    FXMargin := 10;
    FYmargin := 5;
    end;
  {dragging: }
  FHasDragged := false;
  drect := rect(0,0,0,0);
end {Create};


destructor TxyGraph.Destroy;
var p, p1: TSeries;
    p2, p3: pHookedSeries;
    om: pMark;
begin
  FXAxis.Free;
  FYAxis.Free;
  FYAxis2.Free;
  FDimensions.free;
  FAppearance.free;
  FLegend.free;

  p := FSeries;
  while p <> nil do
    begin
    p1 := p;
    p := p.next;
    p1.free;
    end;

  p2 := FHookedSeries;
  while p2 <> nil do
    begin
    p3 := p2;
    p2 := p2^.next;
    p3^.id.hooks[self] := nil;
    dispose(p3);
    end;

  while FMarks <> nil do
  begin
    om := FMarks;
    FMarks := FMarks^.next;
    Dispose(om);
  end;
  inherited Destroy;
end;

{---------------------------------------------------------------------
    #6b. TxyGraph  - mouse routines
 ---------------------------------------------------------------------}

{general outline of responding to mouse events:
  if it's not the left mouse button, just call the inherited handler

  if the shift button is down, the user wants to change the plot area
  if the ctrl button is down, the user wants to run the designer
  if Draggable is set to true:
     if the mouse is in the top left corner, the user wants to move the graph
     if the mouse is in the bottom right corner, the user wants to resize the graph}

procedure TxyGraph.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
const
  SC_DragMove = $f012;
  SC_SizeLR   = $f008;
begin
  if (Button = mbLeft) then
  begin
    if (ssCtrl in Shift) then
      if assigned(FOnWantDesigner) then
        FOnWantDesigner(self)
      else
    else
    if (ssShift in Shift) then setupdragging(x,y) else
    with FDimensions do
     {3. is mouse in bottom right corner? }
      if FDraggable and (Width - X < FRight) and (Height - Y < FBottom) then
      begin
        Screen.Cursor := crSizeNWSE;
        ReleaseCapture;
        Perform(WM_SysCommand, SC_SizeLR, 0);
        Screen.Cursor := crDefault;
      end
     {is mouse in top left corner? }
      else if FDraggable and (X < FLeft) and (Y < FTop) then
      begin
        Screen.Cursor := crSize;
        ReleaseCapture;
        Perform(WM_SysCommand, SC_DragMove, 0);
        Screen.Cursor := crDefault;
      end
  end;
  inherited MouseDown(Button, Shift, X, Y);
end {MouseDown};

procedure TxyGraph.SetupDragging;
begin
 if (abs(x -(FLegend.Left+3))<3) and (abs(y-(FLegend.top+3))<3) and FLegend.FVisible then
   begin
   FLegDragging := true;
   dx := FLegend.width;
   dy := FLegend.height;
   end
 else
   begin
   fdragging := true;
   dx := x;
   dy := y;
   screen.cursor := crCross;
   if not FHasDragged then
     begin
     FHoldXAS := FXAxis.Autosizing;
     FHoldXmin := FXAxis.Min;
     FHoldXmax := FXAxis.Max;
     FHoldYAS := FYAxis.Autosizing;
     FHoldYmin := FYAxis.Min;
     FHoldYmax := FYAxis.Max;
     FHoldY2AS := FYAxis2.Autosizing;
     FHoldY2min := FYAxis2.Min;
     FHoldY2max := FYAxis2.Max;
     end;
   FHasDragged := true;
   end;
 with FCanvas.pen do
   begin
   Color := Self.color;
   Mode := pmXOr;
   Style := psDot;
   end;
end;

procedure TxyGraph.MouseMove(Shift: TShiftState; X, Y: Integer);
  procedure drawMyRect(r:TRect);
  begin
  FCanvas.Polyline([pt(r.left, r.top),    pt(r.right,r.top),
                    pt(r.right,r.bottom), pt(r.left, r.bottom),
                    pt(r.left, r.top)]);
  end;
begin
 if fLegDragging then
   begin
   DrawMyRect(drect);
   drect.left := x;
   drect.Top := y;
   drect.Right := x+dx;
   drect.Bottom := y+dy;
   DrawMyRect(drect);
   end;
 if fdragging then
 begin
   drawMyRect(drect);
   if x > dx then
   begin
     drect.left := dx;
     drect.right := x;
   end
   else
   begin
     drect.left := x;
     drect.right := dx;
   end;
   if y > dy then
   begin
     drect.top := dy;
     drect.bottom := y;
   end
   else
   begin
     drect.top := y;
     drect.bottom := dy;
   end;
   drawMyRect(drect);
 end;
 inherited mousemove(shift, x, y);
end;

procedure TxyGraph.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  x1,x2,y1,y2,y3,y4: Double;
begin
  If FLegDragging or FDragging then
    begin
    with FCanvas.pen do
      begin
      Color := clBlack;
      Mode := pmCopy; {//pmXOr;}
      end;
    drect := rect(0,0,0,0);
    releaseCapture;
    end;
  if FLegDragging then
    begin
    FLegend.left := x;
    FLegend.top := y;
    FLegDragging := false;
    end;
  if Fdragging then
  begin
    FDragging := False;
    screen.cursor := crDefault;
    if (abs (x - dx) > MouseToleranceMove) and (abs (y - dy) > MouseToleranceMove) then
    begin
      if x > dx then
      begin
        x1 := getmousex(dx);
        x2 := getmousex(x);
      end
      else
      begin
        x1 := getmousex(x);
        x2 := getmousex(dx);
      end;
      if y < dy then
      begin
        y1 := getmousey(dy);
        y2 := getmousey(y);
        y3 := getmousey2(dy);
        y4 := getmousey2(y);
      end
      else
      begin
        y1 := getmousey(y);
        y2 := getmousey(dy);
        y3 := getmousey2(y);
        y4 := getmousey2(dy);
      end;
      plotting := false;
      {restrain zooming to 1000x original: errors start
       to become visible at extremely high magnifications}
      if abs(x1-x2) < (FHoldXMax - FHoldXMin)/1000 then
        if x1 < x2 then
          x2 := x1 + (FHoldXMax - FHoldXMin)/1000
        else
          x1 := x2 + (FHoldXMax - FHoldXMin)/1000;
      Fxaxis.min := x1;
      Fxaxis.max := x2;
      if abs(y1-y2) < (FHoldYMax - FHoldYMin)/1000 then
        if y1 < y2 then
          y2 := y1 + (FHoldYMax - FHoldYMin)/1000
        else
          y1 := y2 + (FHoldYMax - FHoldYMin)/1000;
      Fyaxis.min := y1;
      Fyaxis.max := y2;
      if abs(y3-y4) < (FHoldY2Max - FHoldY2Min)/1000 then
        if y3 < y4 then
          y4 := y3 + (FHoldY2Max - FHoldY2Min)/1000
        else
          y3 := y4 + (FHoldY2Max - FHoldY2Min)/1000;
      Fyaxis2.min := y3;
      FYAxis2.max := y4;
      plotting := true;
    end
    else
    begin
      FHasDragged := false;
      plotting := false;
      FXaxis.autosizing := FHoldXAS;
      if not FHoldXAS then
      begin
        FXAxis.min := FHoldXMin;
        FXAxis.max := FHoldXMax;
      end;
      FYaxis.autosizing := FHoldYAS;
      if not FHoldYAS then
      begin
        FYAxis.min := FHoldYMin;
        FYAxis.max := FHoldYMax;
      end;
      FYaxis2.autosizing := FHoldY2AS;
      if not FHoldY2AS then
      begin
        FYAxis2.min := FHoldY2Min;
        FYAxis2.max := FHoldY2Max;
      end;
      plotting := true;
    end;
  end;
  inherited mouseup(button, shift, x, y);
end;

{---------------------- Reverse Metric Routines -------------------------------}

function TxyGraph.GetMouseXY(sN, x, y: Integer;
           var index: Integer;
           var xScaled, yScaled, xNearest, yNearest: Double): Boolean;
begin
  xScaled := GetMouseX(x);
  yScaled := GetMouseY(y);
  Result := GetMouseVals(sN, x, index, xNearest, yNearest);
end {GetMouseXY};

{ getmousex/y return the value as determined by the scales }
function TxyGraph.GetMouseX(x: Integer): Double;
begin
  with FXAxis do
    if FLogging then
      Result := FMin*Exp((x - FDimensions.FLeft) / FM)
    else
      Result := FMin + ((x - FDimensions.FLeft) / FM);
end;

function TxyGraph.GetMouseY(y: Integer): Double;
begin
with FYAxis do
  if FLogging then
    Result := FMax*Exp(-(y - FDimensions.FTop) / FM)
  else
    Result := FMax - ((y - FDimensions.FTop) / FM);
end;


function TxyGraph.GetMouseY2(y: Integer): Double;
begin
with FYAxis2 do
  if FLogging then
    Result := FMax*Exp(-(y - FDimensions.FTop) / FM)
  else
    Result := FMax - ((y - FDimensions.FTop) / FM);
end;


{ GetMouseVals returns the value of the data point matching the
mouse point best, or false if the mouse is outside the plot area }
function TxyGraph.GetMouseVals(sN, x: Integer; var index: Integer;
                               var xNearest, yNearest: Double): Boolean;
var
  tx: Double;
{  i: Integer;}
  p: TSeries;
  p1, p2: pgPoint;
begin
  tx := GetMouseX(x);
  p := Series[sN];
  Result := (p <> nil) and (p.FData <> nil) and
                       ((tx <= FXAxis.Max) and (tx >= FXAxis.Min));
  if Result then
  begin
    p1 := p.FData;
{    i := 1;}
    if tx <= p1^.xv then
      p2 := p1
    else
    begin
      while (p1^.next <> nil) and (tx > p1^.xv) do
      begin
        p2 := p1;
        p1 := p1^.next;
{        Inc(i);}
      end;
      if p1^.next = nil then
        p2 := p1
      else if tx = p1^.xv then
        p2 := p1
      else if tx > (p1^.xv + p2^.xv) / 2 then
        p2 := p1
{      else Dec(i);}
    end;
    index := p2^.i;
    xNearest := p2^.xv;
    yNearest := p2^.yv;
  end;
end {GetMouseVals};

{FindNearestPoint - get the handle to the nearest point to the
                    mouse.
  x,y: mouse position.
  series: a series to search, or 0 for all series.
  returns the series in series (hence var pars)
FindNearestPoint finds the point nearest _visually_. (if scales
are logged, it may not be the point closest to the data value
of x,y)

comment: Yes, I know, this is a slow way to do it. You could probably
think of faster ways. But this is enough for me.}

function TxyGraph.findNearestPoint(series, x,y:integer):pgPoint;
var current,currs:longint;
    pp,currpp:pgPoint;
    p:TSeries;
    started:boolean;
     function sqr(w:longint):longint;
     begin
       result := w*w;
     end;
begin
  result := nil;
  started := false;
  p := FSeries;
  currs := 0;
  while (p <> nil) do
    begin
    if (series = 0) or (p.FSeriesindex = series) then
      begin
      pp := p.Data;
      while pp <> nil do
        begin
        with pp^ do
          if not started or (sqr(fx(xv)-x) + sqr(Fy(yv) - y) < current) then
            begin
            currs := p.FSeriesIndex;
            current := sqr(fx(xv)-x) + sqr(Fy(yv) - y);
            currpp := pp;
            started := true;
            end;
        pp := pp^.next;
        end;
      end;
    p := p.next;
    end;
  if started then
    begin
    result := currpp;
    series := currs;
    end;
end;

{---------------------------------------------------------------------
    #6c. TxyGraph  - metrics routines
 ---------------------------------------------------------------------}

procedure TxyGraph.DoRescaleEvent;
begin
  if FPlotting and CheckScales and Assigned(FOnRescale) then
    FOnRescale(self);
end;

{procedure TxyGraph.CalcXMetrics;
begin
  with FXAxis do
  begin
    if (FMin>=FMax) then raise Exception.Create('CalcXMetrics: XMin>=XMax');
    if FLogging then
      FM := (FCurrWidth - FDimensions.FLeft - FDimensions.FRight) / (Ln(FMax) - Ln(FMin))
    else
      FM := (FCurrWidth - FDimensions.FLeft - FDimensions.FRight) / (FMax - FMin);
  end;
end;

procedure TxyGraph.CalcYMetrics;
begin
  with FYAxis do
  begin
    if (FMin>=FMax) then raise Exception.Create('CalcYMetrics: YMin>=YMax');
    if FLogging then
      FM := (FCurrHeight - FDimensions.FTop - FDimensions.FBottom) / (Ln(FMax) - Ln(FMin))
    else
      FM := (FCurrHeight - FDimensions.FTop - FDimensions.FBottom) / (FMax - FMin);
  end;
end;

procedure TxyGraph.CalcY2Metrics;
begin
  with FYAxis2 do
  begin
    if (FMin>=FMax) then raise Exception.Create('CalcY2Metrics: YMin>=YMax');
    if FLogging then
      FM := (FCurrHeight - FDimensions.FTop - FDimensions.FBottom) / (Ln(FMax) - Ln(FMin))
    else
      FM := (FCurrHeight - FDimensions.FTop - FDimensions.FBottom) / (FMax - FMin);
  end;
end;
 }
procedure TxyGraph.CalcMetrics;
begin
  if FXaxis.ScaledOk then FXAxis.CalcMetrics;
  if FYAxis.ScaledOK then FYAxis.CalcMetrics;
  if FYAxis2.scaledOK then FYAxis2.CalcMetrics;
end;

function TxyGraph.fx(v: Double): Integer;
var w: Double;
begin
{  if not FXAxis.ScaledOk then raise exception.create('call to fx when x fails');}
  with FXAxis do
    if FLogging then
      w := (Ln(v) - Ln(FMin)) * FM
    else
      w :=(v - FMin) * FM;
  if abs(w) > 20000 then
    Result := 20000
  else
    Result := Round(w) + FDimensions.FLeft + FXOffset;
{20000  is large compared with device pixel size; this avoids range
 error if some points are far outside the axis range}
end;

function TxyGraph.fy(v: Double): Integer;
var w: Double;
begin
{  if not FYAxis.ScaledOk then raise exception.create('call to fy when y fails');}
  with FYAxis do
    if FLogging then
      w := (Ln(FMax) - Ln(v)) * FM
    else
      w := (FMax - v) * FM;
  if abs(w) > 20000 then
    Result := 20000
  else
    Result := Round(w) + FDimensions.FTop + FYOffset;
end;

function TxyGraph.fy2(v: Double): Integer;
var w: Double;
begin
{  if not FYAxis2.ScaledOk then raise exception.create('call to fy2 when y2 fails');}
  with FYAxis2 do
    if FLogging then
      w := (Ln(FMax) - Ln(v)) * FM
    else
      w := (FMax - v) * FM;
  if abs(w) > 20000 then
    Result := 20000
  else
    Result := Round(w) + FDimensions.FTop + FYOffset;
end;

{---------------------- graph scaling ------------------------------------}
procedure TxyGraph.DoTightFit;
var
  initx,inity,inity2: Boolean;
  p: TSeries;
begin
  p := FSeries;
  while p <> nil do
  begin
    if p.Changed then p.GetMinMax;
    p := p.next;
  end;

  initx := true;
  inity := true;
  inity2 := true;
  p := FSeries;
  while p <> nil do
  begin
    if p.FActive then
    begin
      if initx then
        begin
        FXAxis.oMin := p.xvMin;
        FXAxis.oMax := p.xvMax;
        initx := false;
        end
      else
        begin
        if p.xvMin < FXAxis.oMin then FXAxis.oMin := p.xvMin;
        if p.xvMax > FXAxis.oMax then FXAxis.oMax := p.xvMax;
        end;
      if p.FWhichYAxis = FYAxis then
        begin
        if inity then
          begin
          FYAxis.oMin := p.yvMin;
          FYAxis.oMax := p.yvMax;
          inity := false
          end
        else
          begin
          if p.yvMin < FYAxis.oMin then FYAxis.oMin := p.yvMin;
          if p.yvMax > FYAxis.oMax then FYAxis.oMax := p.yvMax;
          end;
        end
      else
        begin
        if inity2 then
          begin
          FYAxis2.oMin := p.yvMin;
          FYAxis2.oMax := p.yvMax;
          inity2 := false
          end
        else
          begin
          if p.yvMin < FYAxis2.oMin then FYAxis2.oMin := p.yvMin;
          if p.yvMax > FYAxis2.oMax then FYAxis2.oMax := p.yvMax;
          end;
        end
    end;
    p := p.next;
  end;
end {DoTightFit};

(*function TxyGraph.DoResize: Boolean;
var b: Boolean;
begin
  b := FYAxis.DoResize;    {force evaluation of both functions - need side effects!}
  b := FYAxis2.DoResize or b;
  Result := FXAxis.DoResize or b;
end;*)

{---------------------------------------------------------------------
    #6d. TxyGraph  - painting routines
 ---------------------------------------------------------------------}

procedure TxyGraph.ClipGraph;
var
  ClipRgn: HRgn;
begin
   ClipRgn := CreateRectRgn(FDimensions.FLeft + FXOffset,
                            FDimensions.FTop + FYOffset,
                            FCurrWidth - FDimensions.FRight + 1 + FXOffset,
                            FCurrHeight - FDimensions.FBottom + 1 + FYOffset);
   SelectClipRgn(FCanvas.Handle, ClipRgn);
   DeleteObject(ClipRgn);
end;

procedure TxyGraph.UnclipGraph;
var
  ClipRgn: HRgn;
begin
  { note for confused readers: the shortcut to unclip that is valid for
       screen metrics is not valid for printers :) }
   ClipRgn := CreateRectRgn(FXOffset,
                            FYOffset,
                            FCurrWidth + 1 + FXOffset,
                            FCurrHeight + 1 + FYOffset);
   SelectClipRgn(FCanvas.Handle, ClipRgn);
   DeleteObject(ClipRgn);
end;

procedure TxyGraph.DrawLineSegment(x1, y1, x2, y2: Integer);
begin
  FCanvas.MoveTo(x1, y1);
  FCanvas.LineTo(x2, y2);
end;

procedure TxyGraph.DrawXGridlines;
var
  tick, maxTick: Double;
  tx1, ty1, ty2: Word;
  b: Boolean;
  tempLogTickInfo : TLogTickInfo;
begin
  FCanvas.Pen.Color := FAppearance.FGridColor;
  FCanvas.Pen.Style := FAppearance.FGridStyle;
  SetBkColor(FCanvas.handle, FAppearance.FBkgdColor);
  FCanvas.Brush.Color := Color;
  with FXAxis do
    maxTick := FMax + 0.001*(FMax-FMin); { rounding errors might exclude last point }
  with FYAxis do
  begin
    ty1 := fy(FMin);
    ty2 := fy(FMax);
  end;
  with FXAxis do
  begin
    if FLogging then tempLogTickInfo := FLogTickInfo;
    tick := GetFirstTick(tempLogTickInfo);
   {don't draw the first grid line on top of axis:}
    if tick < FMin + 0.01*FStep then
      tick := GetNextTick(tick, tempLogTickInfo, b);
    while tick < maxTick do
    begin
      tx1 := fx(tick);
      DrawLineSegment(tx1, ty1, tx1, ty2);
      tick := GetNextTick(tick, tempLogTickInfo, b);
    end;
  end;
end {DrawXGridlines};

procedure TxyGraph.DrawYGridlines;
var
  tick, maxTick: Double;
  tx1, tx2, ty1: Word;
  b: Boolean;
  tempLogTickInfo : TLogTickInfo;
  wyax:boolean;
begin
  wyax := (cYAxis = FYAxis);
  FCanvas.Pen.Color := FAppearance.FGridColor;
  FCanvas.Pen.Style := FAppearance.FGridStyle;
  SetBkColor(FCanvas.handle, FAppearance.FBkgdColor);
  with cYAxis do
    maxTick := FMax + 0.001*(FMax-FMin); { rounding errors might exclude last point }
  with FXAxis do
  begin
    tx1 := fx(FMin);
    tx2 := fx(FMax);
  end;
  with cYAxis do
  begin
    if FLogging then tempLogTickInfo := FLogTickInfo;
    tick := GetFirstTick(tempLogTickInfo);
   {don't draw the first grid line on top of axis:}
    if tick < FMin + 0.01*FStep then
      tick := GetNextTick(tick, tempLogTickInfo, b);
    while tick < maxTick do
    begin
      if wyax then ty1 := fy(tick) else ty1 := fy2(tick);
      DrawLineSegment(tx1, ty1, tx2, ty1);
      tick := GetNextTick(tick, tempLogTickInfo, b);
    end;
  end;
end {DrawYGridlines};

procedure TxyGraph.DrawXTickMarks;
var
  tick, maxTick: Double;
  tx, ty1, ty2: {Word} longint;
  b: Boolean;
  tempLogTickInfo : TLogTickInfo;
begin
  FCanvas.Pen.Color := FAppearance.FAxesColor;
  FCanvas.Pen.Style := psSolid;
  SetBkColor(FCanvas.handle, FAppearance.FMarginColor);
  with FXAxis do
    maxTick := FMax + 0.001*(FMax-FMin);
  ty1 := base;
  if reverse then
    ty2 := ty1 - FDimensions.FTMLength
  else
    ty2 := ty1 + FDimensions.FTMLength;
  with FXAxis do
  begin
    if FLogging then tempLogTickInfo := FLogTickInfo;
    tick := GetFirstTick(tempLogTickInfo);
    while tick < maxTick do
    begin
      tx := fx(tick);
      DrawLineSegment(tx, ty1, tx, ty2);
      tick := GetNextTick(tick, tempLogTickInfo, b);
    end;
  end;
end {DrawXTickMarks};

procedure TxyGraph.DrawYTickMarks;
var
  tick, maxTick: Double;
  tx1, tx2, ty: {Word} longint;
  b, wyax: Boolean;
  tempLogTickInfo : TLogTickInfo;
begin
  wyax := (cYAxis = FYAxis);
  FCanvas.Pen.Color := FAppearance.FAxesColor;
  FCanvas.Pen.Style := psSolid;
  SetBkColor(FCanvas.handle, FAppearance.FMarginColor);
  with cYAxis do
    maxTick := FMax + 0.001*(FMax-FMin);
  tx1 := base;
  if reverse then
    tx2 := tx1 + FDimensions.FTMLength
  else
    tx2 := tx1 - FDimensions.FTMLength;
  with cYAxis do
  begin
    if FLogging then tempLogTickInfo := FLogTickInfo;
    tick := GetFirstTick(tempLogTickInfo);
    while tick < maxTick do
    begin
      if wyax then ty := fy(tick) else ty := fy2(tick);
      DrawLineSegment(tx1, ty, tx2, ty);
      tick := GetNextTick(tick, tempLogTickInfo, b);
    end;
  end;
end {DrawYTickMarks};

function  TAxis.LabelString(tick: Double): string;
begin
  if not FShowAsTime then
    if (abs(tick) < (FMax - FMin)*0.001 {zero}) or
        ((abs(tick) > 0.000999) and (abs(tick) < 9999)) then
      Result := FloatToStrF(tick, ffFixed, 5, FLabelDec)
    else {very small or very large}
      Result := FloatToStrF(tick, ffExponent, 2, 0)
  else
   result := FormatDateTime(FDateFormat,tick);
end;

function TxyGraph.DrawXLabels;
var
  tick, maxTick: Double;
  { tx, }ty: {Word} longint;
  lblStr: string;
  drawIt: Boolean;
  tempLogTickInfo : TLogTickInfo;

begin
{ X-axis labels }
  FCanvas.Font := FAppearance.FLabelFont;
  FAltCanvas.Font := FAppearance.FLabelFont; {if FIsMetafiling FCanvas <> FAltCanvas}
  FCanvas.Brush.Style := bsClear;
  SetBkColor(FCanvas.handle, FAppearance.FMarginColor);
  with FXAxis do
    maxTick := FMax + 0.001*(FMax-FMin);   { rounding errors might exclude last point }
  if reverse then
    ty := base - (FDimensions.FTMLength + FDimensions.FXAxisLabelDistance)
  else
    ty := base + FDimensions.FTMLength + FDimensions.FXAxisLabelDistance;
  with FXAxis do
  begin
    if FLogging then tempLogTickInfo := FLogTickInfo;
    tick := GetFirstTick(tempLogTickInfo);
    drawIt := true;
    while tick < maxTick do
    begin
      lblStr := LabelString(tick);
      if drawIt or (maxTick - tick < FStep) {ie, last one} then
        if reversed then
          FCanvas.TextOut(fx(tick) - FAltCanvas.TextWidth(lblStr) div 2,
                        ty - FAltCanvas.TextHeight(lblstr),
                        lblStr)
        else
          FCanvas.TextOut(fx(tick) - FAltCanvas.TextWidth(lblStr) div 2,
                        ty,
                        lblStr);
      tick := GetNextTick(tick, tempLogTickInfo, drawIt);
    end;
  end;
  result := FAltcanvas.Textheight(lblstr);
end {DrawXLabels};

function TxyGraph.DrawYLabels;
var
  tick, maxTick: Double;
  tx, ty: Integer;
  lblStr: string;
  drawIt, wyax: Boolean;
  tempLogTickInfo : TLogTickInfo;
begin
{ Y-axis Labels }
  FCanvas.Font := FAppearance.FLabelFont;
  FAltCanvas.Font := FAppearance.FLabelFont; {if FIsMetafiling FCanvas <> FAltCanvas}
  SetBkColor(FCanvas.handle, FAppearance.FMarginColor);
  wyax := (cYAxis = FYAxis);
  with cYAxis do
    maxTick := FMax + 0.001*(FMax-FMin);   { rounding errors might exclude last point }
  if reverse then
    tx := base + (FDimensions.FTMLength + FDimensions.FYAxisLabelDistance)
  else
    tx := base - (FDimensions.FTMLength + FDimensions.FYAxisLabelDistance);
  with cYAxis do
  begin
    if FLogging then tempLogTickInfo := FLogTickInfo;
    tick := GetFirstTick(tempLogTickInfo);
    drawIt := true;
    while tick < maxTick do
    begin
      lblStr := LabelString(tick);
      if wyax then ty := fy(tick) else ty := fy2(tick);
      if drawIt or (maxTick - tick < FStep) then
        if reverse then
          FCanvas.TextOut(tx,
                        ty - FAltCanvas.TextHeight(lblStr) div 2,
                        lblStr)
        else
          FCanvas.TextOut(tx - FAltCanvas.TextWidth(lblStr),
                        ty - FAltCanvas.TextHeight(lblStr) div 2,
                        lblStr);
      tick := GetNextTick(tick, tempLogTickInfo, drawIt);
    end;
  end;
  result := FAltCanvas.TextWidth(lblstr);
end {DrawYLabels};

procedure TxyGraph.DrawGraphTitle;
var
  tx: Word;
begin
  FCanvas.Font := FAppearance.FCaptionFont;
  FAltCanvas.Font := FAppearance.FCaptionFont;
          {if FIsMetafiling FCanvas <> FAltCanvas}
  SetBkColor(FCanvas.handle, FAppearance.FMarginColor);
  with FDimensions, FAppearance do
  begin
    tx := FLeft + (FCurrWidth - FLeft - FRight) div 2 + FXOffset;
    with FAltCanvas do
      FCanvas.TextOut(tx - TextWidth(FGraphTitle) div 2,
              FTop - FGraphTitleDistance - TextHeight(FGraphTitle) + FYOffset,
              FGraphTitle);
  end;
end {DrawGraphTitle};

procedure TXyGraph.DrawMarks;

  function inrange(v,min,max:double):boolean;
  begin
   result := (v >= min) and (v <= max);
  end;

  procedure MarkFontOut(x,y:integer;angle:double;
               s:string; offset:integer; color:TColor);
  var LogRec  : TLOGFONT;     {* Storage area for font information *}
      OldFont,
      NewFont: HFONT;
      xoffst,yoffst:integer;
{      c:Tcolor;}
  begin
    FCanvas.Font := FAppearance.FCaptionFont;
    { Get the current font information. We only want to modify the angle }
{    c := FCanvas.Font.Color;}
    FCanvas.Font.Color := color;
    GetObject(FCanvas.Font.Handle,SizeOf(LogRec),@LogRec);
    LogRec.lfEscapement := trunc(angle *10 * 180 / pi);
    LogRec.lfOrientation := trunc(angle *10 * 180 / pi);    {see win32 api?}
    LogRec.lfOutPrecision := OUT_TT_ONLY_PRECIS;  {doesn't work under win95??}
    NewFont := CreateFontIndirect(LogRec);
    OldFont := SelectObject(Canvas.Handle,NewFont);  {Save old font}
  { offsets }
    xoffst := Round(FScale*(sin(angle) * offset));
    yoffst := Round(FScale*(cos(angle) * offset));
    FCanvas.TextOut(x - xoffst, y - yoffst, s);
    NewFont := SelectObject(Canvas.Handle,OldFont); {Restore oldfont}
    DeleteObject(NewFont);
{    FCanvas.font.color := c;  }
  end {markFontOut};

  procedure drawxmark(p:Pmark);
  begin
    Fcanvas.pen.color := p^.color;
    DrawLineSegment(fx(p^.x1), fy(FYAxis.FMin), fx(p^.x1), fy(FYAxis.FMax));
    MarkFontOut(fx(p^.x1),fy(FYAxis.FMin) - 3, pi/2, p^.caption, FCanvas.font.size + 5,
                p^.color);
  end;

  procedure drawymark(p:Pmark);
 { var c:Tcolor;}
  begin
    Fcanvas.pen.color := p^.color;
    DrawLineSegment(fx(FXAxis.Fmin), fy(P^.y1), fx(FXAxis.Fmax), fy(p^.y1));
    FCanvas.Font := FAppearance.FCaptionFont;
{    c := FCanvas.font.color;}
    FCanvas.font.color := p^.color;
    FCanvas.TextOut(FDimensions.FLeft + Round(FScale*10) + FXOffset,
                    fy(p^.y1) - Round(FScale*2) - FAltCanvas.TextHeight(p^.caption),
                    p^.caption);
{    FCanvas.font.color := c;}
  end;

  procedure drawboxmark(p:Pmark);
  var d:integer;
{      c:Tcolor;}
  begin
    d := trunc(FScale*p^.x2);
    FCanvas.pen.color := p^.color;
    FCanvas.brush.style := bsclear;
    FCanvas.rectangle(fx(p^.x1) - d, fy(p^.y1) - d,
                      fx(p^.x1) + d, fy(p^.y1) + d);
 {   c := FCanvas.font.color;}
    FCanvas.font.color := p^.color;
    FCanvas.TextOut(fx(p^.x1) - d, fy(p^.y1) + d, p^.caption);
{    FCanvas.font.color := c;  }
  end;

  procedure drawLineMark(p:Pmark);

    procedure swap(var n1,n2:integer);
    var t:integer;
    begin
      t := n1;
      n1 := n2;
      n2 := t;
    end;

  var minx,miny,maxy,maxx:integer;
      angle:double;
  begin {drawLineMark}
    FCanvas.pen.color := p^.color;
    minx := fx(p^.x1);
    miny := fy(p^.y1);
    maxx := fx(p^.x2);
    maxy := fy(p^.y2);
    if maxx < minx then
    begin
      swap(minx,maxx);
      swap(miny,maxy);
    end;
    DrawLineSegment(minx, miny, maxx, maxy);
    if maxx - minx = 0 then
      angle := pi/2
    else
      angle := - arctan((maxy-miny)/(maxx-minx));
      MarkFontOut(minx,miny,angle,p^.caption, FCanvas.font.size + 5, p^.color);
  end {drawLineMark};

var p: pmark;
begin {DrawMarks}
  if not FAppearance.FAllowedOut then
    ClipGraph;
  p := fmarks;
  while p <> nil do
  begin
    if p^.status = when then
      case p^.marktype of
      mtXMark: drawxmark(p);
      mtYMark: drawymark(p);
      mtPoint: drawboxmark(p);
      mtLine : drawLineMark(p);
      end;
    p := p^.next;
  end;
 if not FAppearance.FAllowedOut then
   UnclipGraph;
end {DrawMarks};


procedure TxyGraph.ColorBackground;
begin
  FCanvas.brush.color := FAppearance.FMarginColor;
  FCanvas.brush.style := bsSolid;
  FCanvas.pen.color := FAppearance.FMarginColor;
  FCanvas.pen.style := psSolid;
  SetBkColor(FCanvas.handle, FAppearance.FMarginColor);
  With FDimensions do
   FCanvas.rectangle(0 + FXOffset, 0 + FYOffset,
                     FCurrWidth + FXOffset,
                     FCurrHeight + FYOffSet);
  FCanvas.brush.color := FAppearance.FBkgdColor;
  FCanvas.brush.style := bsSolid;
  FCanvas.pen.color := FAppearance.FBkgdColor;
  FCanvas.pen.style := psSolid;
  SetBkColor(FCanvas.handle, FAppearance.FBkgdColor);
  With FDimensions do
   FCanvas.rectangle(Fleft + FXOffset, Ftop + FYOffSet,
                     FCurrWidth - (FRight) + FXOffset,
                     FCurrHeight - (FBottom) + FYOffSet);
     {same as (fx(FXAxis.FMin), fy(FYAxis.FMax), fx(FXAxis.FMax),
      fy(FYAxis.FMin)) , except that fx/fy may not be available}                     ;
   { FCanvas.brush.color := Color;}
end {ColorBackground};

procedure TxyGraph.DrawXAxis;
var
  holdXStep: Double;
  ypoint,tx,ty,ts, oldPenWidth:integer;
  s:string;
begin
 ts := 0;
 with FAppearance, FXAxis do
  begin
  if FYAxis.ScaledOK then
    case FOffsetType of
     ao_Minimum:ypoint := fy(FYAxis.FMin);
     ao_Maximum:ypoint := fy(FYAxis.FMax);
     ao_percentage:ypoint := fy(FYAxis.FMin +
                (FOffset / 100 * (FYAxis.FMax - FYAxis.FMin)));
     ao_absolute:ypoint := fy(FOffset);
     else ypoint := fy(FYAxis.FMin);
    end
  else
    with FDimensions do ypoint := FCurrHeight - (FBottom) + FYOffSet;
  if ScaledOK then
    begin
    holdXStep := FStep;
    {if the number of steps is ridiculously large, do something reasonable:}
    if FShowAsTime then
      begin
       FStep := GetDatestep;
       while ((FMax - FMin) / FStep > FMaxSteps) do
        begin
        inc(FDateTickType); {this can't finish up in an infinite loop
           because the last datetick type will always terminate the loop}
        FStep := GetDatestep;
        SetDateMinMax;
        {FGraph.}Calcmetrics;
        end;
      end else
    if not FLogging and ((FStep = 0) or ((FMax - FMin) / FStep > FMaxSteps)) then
      FStep := GetStep;
    if FLogging then InitLogTicks;
    OldPenWidth := FCanvas.Pen.Width;
    if OldPenWidth > 1 then
      if FAppearance.FPrintLineStyle then FCanvas.Pen.Width := 1 else
       if OldPenWidth > 3 then FCanvas.Pen.Width := oldpenwidth div 2 + 1;
    if not FNoGridlines and FGridlines and FYAxis.ScaledOK then DrawXGridlines;
    if not FShowAxis then
      begin
      FCanvas.Pen.Width := OldPenWidth;
      exit;
      end;
    if FTickMarks then DrawXTickMarks(ypoint,Freversed);
    if FLabelGraph then ts := DrawXLabels(ypoint,Freversed);
    FStep := holdXStep;
    FCanvas.Pen.Color := FAxesColor;
    FCanvas.Pen.Style := psSolid;
    FCanvas.Pen.Width := OldPenWidth;
    DrawLineSegment(fx(FXAxis.FMin), ypoint, fx(FXAxis.FMax), ypoint);
    s := FXAxis.Title;
    end
  else s := ErrorCaption;
  if not FShowAxis then exit;
  SetBkColor(FCanvas.handle, FAppearance.FMarginColor);
  FCanvas.Font := FAppearance.FTitleFont;
  FAltCanvas.Font := FAppearance.FTitleFont; {if FIsMetafiling FCanvas <> FAltCanvas}
  with FDimensions do
    begin
    tx := FLeft + (FCurrWidth - FLeft - FRight) div 2 + FXOffset;
    if FXAxis.freversed then
      ty := ypoint - (FTMLength + FXAxisTitleDistance)
               - FAltCanvas.TextHeight(s) - ts
    else
      ty := ypoint + FTMLength + FXAxisTitleDistance +
          ts {FAltCanvas.TextHeight(FXAxis.Title)};
    end;
  FCanvas.TextOut(tx - FAltCanvas.TextWidth(FXAxis.Title) div 2,
                  ty, s);
  end;
end {DrawXAxis};

procedure TxyGraph.DrawYAxis;
var holdYStep: Double;
    tx,ty,xpoint,ts, oldPenWidth:integer;
    Angle:integer;
    LogRec: TLOGFONT;              { Storage area for font information }
    OldFont, NewFont: HFONT;
    s:string;
begin
 ts := 0;
 with FAppearance, FYAxis do
  begin
  if FXAxis.ScaledOK then
    case FOffsetType of
     ao_Minimum:xpoint := fx(FXAxis.FMin);
     ao_Maximum:xpoint := fx(FXAxis.FMax);
     ao_percentage:xpoint := fx(FXAxis.FMin +
                (FOffset / 100 * (FXAxis.FMax - FXAxis.FMin)));
     ao_absolute:xpoint := fx(FOffset);
     else xpoint :=  fx(FXAxis.FMin);
    end
  else
    with FDimensions do xpoint := FLeft + FXOffSet;
  if ScaledOK then
    begin
    holdYStep := FStep;
    if FShowAsTime then
      begin
       FStep := GetDatestep;
       while ((FMax - FMin) / FStep > FMaxSteps) do
        begin
        inc(FDateTickType);  {this can't finsh up in an infinite loop
           because the last datetick type will always terminate the loop}
        FStep := GetDatestep;
        SetDateMinMax;
        {FGraph.}Calcmetrics;
        end;
      end else
    if not FLogging and ((FStep = 0) or ((FMax - FMin) / FStep > FMaxSteps)) then
      FStep := GetStep;
    if FLogging then FYAxis.InitLogTicks;
    OldPenWidth := FCanvas.Pen.Width;
    if OldPenWidth > 1 then
      if FAppearance.FPrintLineStyle then FCanvas.Pen.Width := 1 else
       if OldPenWidth > 3 then FCanvas.Pen.Width := oldpenwidth div 2 + 1;
    if not FNoGridlines and FGridlines and FXAxis.ScaledOK then DrawYGridlines(FYAxis);
    if not FShowAxis then
      begin
      FCanvas.Pen.Width := OldPenWidth;
      exit;
      end;
    if FTickMarks then DrawYTickMarks(FYAxis, xpoint,Freversed);
    if FLabelGraph then ts := DrawYLabels(FYAxis, xpoint,Freversed);
    FStep := holdYStep;
    FCanvas.Pen.Color := FAxesColor;
    FCanvas.Pen.Style := psSolid;
    FCanvas.Pen.Width := OldPenWidth;
    DrawLineSegment(xpoint, fy(FYAxis.FMax), xpoint, fy(FYAxis.FMin));
    s := FYAxis.Title;
    end
  else s := ErrorCaption;
{Y-axis title}
  if not FShowAxis then exit;
  SetBkColor(FCanvas.handle, FAppearance.FMarginColor);
  FCanvas.Font := FAppearance.FTitleFont;
  FAltCanvas.Font := FAppearance.FTitleFont; {if FIsMetafiling FCanvas <> FAltCanvas}
  Angle := 90;
  { Get the current font information. We only want to modify the angle }
  GetObject(FCanvas.Font.Handle, SizeOf(LogRec), @LogRec);
  LogRec.lfEscapement := Angle * 10;
  LogRec.lfOrientation := Angle * 10; {see win32 api?}
  LogRec.lfOutPrecision := OUT_TT_ONLY_PRECIS;
       {The microsoft documentation claims that this forces use of
        a truetype font (the only rotatable fonts commonly available).
         -> This does not work under win95. Choose a truetype font
              for the Titlefont property instead for vertical text
         -> Works as claimed under winNT (4.0)
        You should use truetype fonts anyway for printing}
  NewFont := CreateFontIndirect(LogRec);
  OldFont := SelectObject(FCanvas.Handle, NewFont); {Save old font}
  with FDimensions do
  begin
    if Freversed then
      tx := xpoint + (FTMLength + FYAxisTitleDistance)
    else
      tx := xpoint - (FTMLength + FYAxisTitleDistance);
    ty := FTop + (FCurrHeight - FTop - FBottom) div 2 + FYOffset;
  end;
  with FAltCanvas do
  begin
    if Freversed then
      tx := tx + ts
    else
      tx := tx - ts - TextHeight(s);
    if tx < 0 then tx := 0;
    FCanvas.TextOut(tx, ty + (TextWidth(s) div 2), s);
  end;
  NewFont := SelectObject(FCanvas.Handle, OldFont); {Restore oldfont}
  DeleteObject(NewFont);
  end;
end {DrawYAxis};

procedure TxyGraph.DrawY2Axis;
var holdYStep: Double;
    tx,ty,xpoint,ts, OldPenWidth:integer;
    Angle:integer;
    LogRec: TLOGFONT;              { Storage area for font information }
    OldFont, NewFont: HFONT;
    s:string;
begin
 ts := 0;
 with FAppearance, FYAxis2 do
  begin
  if FXAxis.ScaledOK then
    case FOffsetType of
       ao_Minimum:xpoint := fx(FXAxis.FMin);
       ao_Maximum:xpoint := fx(FXAxis.FMax);
       ao_percentage:xpoint := fx(FXAxis.FMin +
                  (FOffset / 100 * (FXAxis.FMax - FXAxis.FMin)));
       ao_absolute:xpoint := fx(FOffset);
    end
  else
    with FDimensions do xpoint := FCurrWidth - (FRight) + FXOffSet;
  if ScaledOK then
    begin
    holdYStep := FStep;
    if FShowAsTime then
      begin
       FStep := GetDatestep;
       while ((FMax - FMin) / FStep > FMaxSteps) do
        begin
        inc(FDateTickType);  {this can't finish up in an infinite loop
           because the last datetick type will always terminate the loop}
        FStep := GetDatestep;
        SetDateMinMax;
        {FGraph.}Calcmetrics;
        end;
      end else
    if not FLogging and ((FStep = 0) or ((FMax - FMin) / FStep > FMaxSteps)) then
      FStep := GetStep;
    if FLogging then InitLogTicks;
    OldPenWidth := FCanvas.Pen.Width;
    if OldPenWidth > 1 then
      if FAppearance.FPrintLineStyle then FCanvas.Pen.Width := 1 else
       if OldPenWidth > 3 then FCanvas.Pen.Width := oldpenwidth div 2 + 1;
    if not FNoGridlines and FGridlines and FXAxis.scaledOK then DrawYGridlines(FYAxis2);
    if not FShowAxis then
      begin
      FCanvas.Pen.Width := OldPenWidth;
      exit;
      end;
    if FTickMarks then DrawYTickMarks(FYAxis2, xpoint,Freversed);
    SetBkColor(FCanvas.Handle,Color);
    if FLabelGraph then ts := DrawYLabels(FYAxis2, xpoint,Freversed);
    FStep := holdYStep;
    FCanvas.Pen.Color := FAxesColor;
    FCanvas.Pen.Style := psSolid;
    FCanvas.Pen.Width := OldPenWidth;
    DrawLineSegment(xpoint, fy2(FYAxis2.FMax), xpoint, fy2(FYAxis2.FMin));
    s := Title;
    end
  else s := ErrorCaption;
{Y2-axis title}
  if not FShowAxis then exit;
  SetBkColor(FCanvas.handle, FAppearance.FMarginColor);
  FCanvas.Font := FAppearance.FTitleFont;
  FAltCanvas.Font := FAppearance.FTitleFont; {if FIsMetafiling FCanvas <> FAltCanvas}
  Angle := 90;
  { Get the current font information. We only want to modify the angle }
  GetObject(FCanvas.Font.Handle, SizeOf(LogRec), @LogRec);
  LogRec.lfEscapement := Angle * 10;
  LogRec.lfOrientation := Angle * 10; {see win32 api?}
  LogRec.lfOutPrecision := OUT_TT_ONLY_PRECIS;
       {The microsoft documentation claims that this forces use of
        a truetype font (the only rotatable fonts commonly available).
         -> This does not work under win95. Choose a truetype font
              for the Titlefont property instead for vertical text
         -> Works as claimed under winNT (4.0)
        You should use truetype fonts anyway for printing}
  NewFont := CreateFontIndirect(LogRec);
  OldFont := SelectObject(FCanvas.Handle, NewFont); {Save old font}
  with FDimensions do
  begin
    if Freversed then
      tx := xpoint + (FTMLength + FYAxisTitleDistance)
    else
      tx := xpoint - (FTMLength + FYAxisTitleDistance);
    ty := FTop + (FCurrHeight - FTop - FBottom) div 2 + FYOffset;
  end;
  with FAltCanvas do
  begin
    if Freversed then
      tx := tx + ts
    else
      tx := tx - ts - TextHeight(s);
    if tx < 0 then tx := 0;
    FCanvas.TextOut(tx, ty + (TextWidth(s) div 2), s);
  end;
  NewFont := SelectObject(FCanvas.Handle, OldFont); {Restore oldfont}
  DeleteObject(NewFont);
  end;
end {DrawY2Axis};

procedure TxyGraph.DrawAxes;
begin
  if (FCanvas <> Printer.Canvas) or FAppearance.FBkgdWhenPrint then
    ColorBackground;
  DrawXAxis;
  DrawYAxis;
  DrawY2Axis;
  if FAppearance.FLabelGraph then DrawGraphTitle;
  SetBkColor(FCanvas.Handle,FAppearance.FBkgdColor);
end;

procedure TxyGraph.DrawFunction(F: PlotFunction; var parms;
                x1, x2: Double;
                color: TColor; style: TPenStyle; steps: Word);
var
  x, step: Double;
  j: Word;
  started: Boolean;
  holdColor: TColor;
  holdStyle: TPenStyle;
begin
  holdColor := FCanvas.Pen.Color;
  holdStyle := FCanvas.Pen.Style;
  FCanvas.Pen.Color := color;
  FCanvas.Pen.Style := style;
  with FXAxis do
    if x1 = x2 then
    begin
      x1 := FMin;
      x2 := FMax;
    end
    else
    begin
      if x1 < FMin then
        x1 := FMin;
      if x2 > FMax then
        x2 := FMax;
    end;
  if (steps <= 0) or (steps > 3500){best likely printer resolution?} then
  begin
  {this can be made more intelligent, but this is usually smooth enough:}
    step := (x2 - x1) / 100;
    steps := 100;
  end
  else
    step := (x2 - x1) / steps;

  x := x1;
  started := false;
  j := 0;
  try
    if not FAppearance.FAllowedOut then
      ClipGraph;
    while j < steps do begin
      while (not started) and (j < steps) do
        try
          FCanvas.MoveTo( fx(x), fy(F(x, parms)) );
          started := true;
        except
          x := x + step;
          Inc(j);
        end;
      while (started) and (j < steps) do
      try
        x := x + step;
        FCanvas.LineTo( fx(x), fy(F(x, parms)) );
        Inc(j);
      except
        Started := false; {skip out-of-support x's}
      end;
    end;
  finally
    FCanvas.Pen.Color := holdColor;
    FCanvas.Pen.Style := holdStyle;
    if not FAppearance.FAllowedOut then
      UnclipGraph;
  end;
end;

procedure TxyGraph.PaintCross;
var crslen:word;
begin
  if not FAppearance.FAllowedout and
   ((FXAxis.Min > 0) or (FXAxis.Max < 0) or (FYAxis.Min > 0) or (YAxis.Max < 0))
   then exit;
  crslen := round(Fappearance.FCrosslength*Fscale);
  FCanvas.pen.color := FAppearance.FCrossColor;
  DrawLineSegment(fx(0), fy(0) + crslen, fx(0), fy(0) - Crslen - 1);
  DrawLineSegment(fx(0) - Crslen, fy(0), fx(0) + Crslen + 1, fy(0))
end;

procedure TxyGraph.DrawLegend;
var EntryCount,i:word;
    p:TSeries;
    xbase,ybase,xwid,yhei,xleft,ytop,xoffset,yoffset,th,symspace:integer;
    OldPenWidth:word;
 procedure DrawLegendEntry(SeriesName:string;
               LegendStatus: TLegendStatus; LineColor, PointColor:Tcolor;
               DrawLine, DrawPoints: boolean; Pointshape: et_pointshape;
               Linestyle:TPenstyle; Psize, x, y:integer);
 begin
   FCanvas.Brush.Style := bsClear;
   FCanvas.Textout(x+SymSpace+2,y,SeriesName);
   y := y+th div 2;
   if Drawline and ((LegendStatus = lsAll) or (legendstatus = lslines)) then
     with FCanvas do
     begin
     Pen.Style := Linestyle;
     pen.color := LineColor;
     DrawLineSegment(x,y,x+SymSpace,y);
     end;
   if Drawpoints and ((LegendStatus = lsAll) or (legendstatus = lspoints)) then
     with FCanvas do
     begin
     x := x + SymSpace div 2;
     brush.color := pointcolor;
     pen.color := pointcolor;
     case Pointshape of
       ps_Square:Polygon([Pt(x - pSize, y - pSize), Pt(x + psize, y - pSize),
                          Pt(x + pSize, y + psize), Pt(x - pSize, y + pSize)]);
       ps_Circle:Ellipse(x - pSize, y - pSize, x + psize, y + psize);
       ps_Diamond:Polygon([Pt(x, y - pSize), Pt(x + psize, y),
                             Pt(x, y + psize), Pt(x - pSize, y)]);
       ps_cross:begin
                DrawLineSegment(x - pSize, y, x + psize, y);
                DrawLineSegment(x, y - pSize, x, y + psize);
                end;
      end;
      end;
 end;

begin
 EntryCount := 0;
 p := FSeries;
 while p <> nil do
 begin
   if p.FLegendStatus <> lsNone then inc(EntryCount);
   p := p.next;
 end;
 if EntryCount = 0 then exit;
 With Flegend, FCanvas do
   begin
   Font := FFont;
   FAltCanvas.font := FFont;
   Brush.Color := FColor;
   if FborderStyle = bsSingle then
      pen.color := clblack
   else
      pen.color := FColor;
   xleft := round(left*Fscale)+FXoffset;
   ytop := round(top*FScale) + FYOffset;
   xwid := round(width *FScale);
   yhei := round(height*Fscale);
   Rectangle(xleft,ytop,xleft+xwid,ytop+yhei);
   xbase := round((Xmargin+left)*Fscale)+FXoffset;
   ybase := round((top+ymargin)*FScale) +FYOffset;
   symspace := round(FSymbolSpace*Fscale);
   th := FAltCanvas.TextHeight('T');
   if FLayout = lsAcross then
     begin
     yoffset := 0;
     xoffset := (xwid - (fxmargin*2)) div EntryCount;
     end
   else
     begin
     xoffset := 0;
     yoffset := (yhei - (fymargin*2)) div entrycount;
     end;
   end;
 OldPenWidth := FCanvas.Pen.Width;
 if FAppearance.FPrintLineStyle then
       FCanvas.Pen.width := 1;
 i := 0;
 p := FSeries;
 while (p <> nil) do
   begin
   if p.FLegendStatus <> lsNone then
     with p do
     begin
     DrawLegendEntry(SeriesName, LegendStatus, LineColor, PointColor,
         DrawLine, DrawPoints, Pointshape, Linestyle,
         round((Pointsize div 2)*FScale), xbase+i*xoffset,
         ybase + i*yoffset);
     FCanvas.Brush.Color := FLegend.Fcolor;
     inc(i);
     end;
   p := p.next;
   end;
 FCanvas.Pen.Width := OldPenWidth;
end;

function TxyGraph.CheckScalesOK: Boolean;
var
  XOK, YOK, Y2OK, xResized, yResized, y2Resized: Boolean;
begin
  xResized := false;
  with FXAxis do
    begin
    if FAutoSizing then
      xResized := DoResize;
    XOK := CheckScale;
    if XOK and FAutoStepping then
      begin
      if not FLogging then
        FStep := GetStep
      else
        begin
        AdjustLogTickCount;
        CheckDrawMinorLabels;
        end;
      end;
    end;
  yResized := false;
  with FYAxis do
    begin
    if FAutoSizing then
      yResized := DoResize;
    YOK := CheckScale;
    if YOK and FAutoStepping then
      begin
      if not FLogging then
        FStep := GetStep
      else
        begin
        AdjustLogTickCount;
        CheckDrawMinorLabels;
        end;
      end;
    end;
  y2Resized := false;
  with FYAxis2 do
    begin
    if FAutoSizing then
      y2Resized := DoResize;
    Y2OK := CheckScale;
    if Y2OK and FAutoStepping then
      begin
      if not FLogging then
        FStep := GetStep
      else
        begin
        AdjustLogTickCount;
        CheckDrawMinorLabels;
        end;
      end;
    end;

  result := XOK and YOK and Y2OK;
  if result and (xResized or yResized or y2resized) then DoRescaleEvent;
end;

procedure TxyGraph.PaintGraph;
var
  p: TSeries;
  oldPenWidth:word;
  obc:Tcolor;
begin
  if FPlotting and DataFound(nil) then DoTightFit;
  AllScalesOK := CheckScalesOK;
  CalcMetrics;
  obc := SetBkColor(FCanvas.handle, FAppearance.FBkgdColor);
  DrawAxes;
  if not AllScalesOK then exit;
  if Fappearance.FCrossAtZero and not FXaxis.FLogging and
    not FYAxis.Flogging then PaintCross;
  if not FPlotting then Exit;
  {Printing: linewidth for series, points thinner than for axes}
  if FCanvas.Pen.Width >= 3 then
    FCanvas.Pen.Width := (FCanvas.Pen.Width + 1) div 2;
  if FAppearance.FShowMarks then DrawMarks(drawbefore);
  OldPenWidth := FCanvas.Pen.Width;
  if FAppearance.FPrintLineStyle then
       FCanvas.Pen.width := 1;
  p := FSeries;
  while p <> nil do
  begin
    p.Draw;
    p := p.next;
  end;
  FCanvas.Pen.Width := OldPenWidth;
  if FAppearance.FShowMarks then DrawMarks(drawafter);
  if FLegend.FVisible then DrawLegend;
  If not (csDesigning in ComponentState) and Assigned(FOnPaintEnd) then
     FOnPaintEnd(self, FCanvas);
  SetBkColor(FCanvas.handle, obc);
end;

procedure TxyGraph.Paint;
begin
  inherited Paint;
  if FPainting then exit;
    {stop re-entrancy problems if a window - such as a print dialog
     box - is cleared just before printing, calling a paint and
     screwing the printer metrics up}
  FCurrWidth := width;
  FCurrHeight := height;
  PaintGraph;
end;

procedure TxyGraph.Print;
begin
  Printer.BeginDoc;
  PrintOnPage;
  Printer.EndDoc;
end;

procedure TxyGraph.PrintOnPage;
var
  scaleX, scaleY, FontScaleFactor: Double;
  saveMargLeft, saveMargRight, saveMargTop, saveMargBottom,
  saveTMLength, saveXLabelDistance, saveYLabelDistance,
  saveXAxisDistance, saveYAxisDistance, saveGraphTitleDistance: Integer;
  videoFontColor: TColor;
  CapFontSize, titlefontsize, legendfontsize, labelfontsize: Integer;
  videoFontName: string;
  videoFontStyles: TFontStyles;
  videoPenWidth: Integer;
  VertPixPerInch: Integer;
  oldcursor:TCursor;
  s:string;
begin
 FPainting := true;
 try
  oldcursor := screen.cursor;
  Screen.cursor := crHourglass;
  with FDimensions do
  begin
    saveMargLeft := FLeft;
    saveMargRight := FRight;
    saveMargTop := FTop;
    saveMargBottom := FBottom;
    saveTMLength := FTMLength;
    saveXLabelDistance := FXAxisLabelDistance;
    saveYLabelDistance := FYAxisLabelDistance;
    saveXAxisDistance := FXAxisTitleDistance;
    saveYAxisDistance := FYAxisTitleDistance;
    saveGraphTitleDistance := FGraphTitleDistance;
  end;
  videoFontName := Font.Name;
  videoFontColor := Font.Color;
  videoFontStyles := Font.Style;
  videoPenWidth := FCanvas.Pen.Width;
  CapFontSize := FAppearance.FCaptionFont.size;
  titlefontsize := FAppearance.FTitleFont.size;
  legendfontsize := FLegend.font.size;
  labelfontsize := FAppearance.FLabelFont.size;
 {scale & offsets}
  scaleX := Printer.PageWidth / ClientWidth;
  scaleY := Printer.PageHeight / ClientHeight;
  with FDimensions do
  begin
    if scaleX < scaleY then
    begin
      FScale := FScalePct/100 * scaleX;
      if FScalePct + FXOffsetPct > 100 then
        FXOffsetPct := (100 - FScalePct) div 2;
      if FScalePct + FYOffsetPct > 100*(scaleY/scaleX) then
        FYOffsetPct := (Round(100*(scaleY/scaleX)) - FScalePct) div 2;
    end
    else
    begin
      FScale := FScalePct * scaleY / 100;
      if FScalePct + FYOffsetPct > 100 then
        FYOffsetPct := (100 - FScalePct) div 2;
      if FScalePct + FXOffsetPct > 100*(scaleX/scaleY) then
        FXOffsetPct := (Round(100*(scaleX/scaleY)) - FScalePct) div 2;
    end;
    FXOffset := Round(FXOffsetPct * (Printer.PageWidth / 100));
    FYOffset := Round(FYOffsetPct * (Printer.PageHeight / 100));
  end;
  FCurrWidth := Round(FScale*Width);
  FCurrHeight := Round(FScale*Height);
  with FDimensions do
  begin
    FLeft := Round(FScale*FLeft);
    FRight := Round(FScale*FRight);
    FTop := Round(FScale*FTop);
    FBottom := Round(FScale*FBottom);
    FTMLength := Round(FScale*FTMLength);
    FXAxisLabelDistance := Round(FScale*FXAxisLabelDistance);
    FYAxisLabelDistance := Round(FScale*FYAxisLabelDistance);
    FXAxisTitleDistance := Round(FScale*FXAxisTitleDistance);
    FYAxisTitleDistance := Round(FScale*FYAxisTitleDistance);
    FGraphTitleDistance := Round(FScale*FGraphTitleDistance);
  end;
 {set up printer canvas}

  { the requirement for the following line was reported by sviolante@era.co.uk
    this may be the source of the problems many people have been having. This
    would only be required if the printer canvas was not properly initialisd.
    it makes no difference on any of my systems, but I've never had any
    problems with printing. please report if you continue to have problems
    with this line present}
  Printer.Canvas.Font.PixelsPerInch := GetDeviceCaps (Printer.Handle, LOGPIXELSY);

  FCanvas := Printer.Canvas;
  FAltCanvas := Printer.canvas;
  with FCanvas do   {for now there are no special printer font properties}
  begin
    Font.Name := videoFontName;
    Font.Color := videoFontColor;
    Font.Style := videoFontStyles;
    VertPixPerInch := GetDeviceCaps(Printer.Handle, LogPixelsY);
    FontScaleFactor := (FScale / VertPixPerInch) * Canvas.Font.PixelsPerInch;
    Pen.Width := Round(videoPenWidth * FScale);
  end;
  FAppearance.FCaptionFont.size := Round(FontScaleFactor * CapFontSize);
  FAppearance.FTitleFont.size := Round(FontScaleFactor * titlefontsize);
  FLegend.font.size := Round(FontScaleFactor * legendfontsize);
  FAppearance.FLabelFont.size := Round(FontScaleFactor * labelfontsize);

 {print and restore canvas}
  try
    PaintGraph;
    try
      If assigned(FOnPaintEnd) then FOnPaintEnd(self, FCanvas);
    except
    end;
  finally
    FScale := 1.0;
    FXOffset := 0;
    FYOffset := 0;
    FAppearance.FCaptionFont.size := CapFontSize;
    FAppearance.FTitleFont.size := titlefontsize;
    FLegend.font.size := legendfontsize;
    FAppearance.FLabelFont.size := labelfontsize;
    FCanvas := Canvas;
    FAltCanvas := Canvas;

    with FDimensions do
    begin
      FLeft := saveMargLeft;
      FRight := saveMargRight;
      FTop := saveMargTop;
      FBottom := saveMargBottom;
      FTMLength := saveTMLength;
      FXAxisLabelDistance := saveXLabelDistance;
      FYAxisLabelDistance := saveYLabelDistance;
      FXAxisTitleDistance := saveXAxisDistance;
      FYAxisTitleDistance := saveYAxisDistance;
      FGraphTitleDistance := saveGraphTitleDistance;
      screen.cursor := oldcursor;
    end;
  end;
 finally
  FPainting := false;
 end;
end {PrintOnPage};

{fill metafile requires a valid TMetafile to be supplied to it -
 resource allocation/deallocation issues. The internally created
 metafile is produced first, and then copied into the external one}
procedure TxyGraph.FillMetafile;
{thanks to the following people for their help with the Metafiles
 routine: Frank Schmidt, John Biddiscombe, Hendrik Levsen }
var
   {$IFDEF WIN32}
    metafile   : TMetaFile;
    metaCanvas : TMetaFileCanvas;
   {$ELSE}
    Metafile: TDrawMetafile;
    saveWidth, saveHeight: Integer;
   {$ENDIF}
begin
 {$IFDEF WIN32}
  metafile := TMetaFile.Create;
  metafile.Height := height;
  metafile.width := width;
  FCurrWidth := width;
  FCurrHeight := height;
  metaCanvas := TMetafileCanvas.Create(metafile, 0);
  Fcanvas := metaCanvas;
 {$ELSE}
  saveWidth := Width;
  saveHeight := Height;
  Metafile := TDrawMetafile.Create(Width, Height);
  FCanvas:= Metafile.Canvas;
 {$ENDIF}
  try
    PaintGraph;
    try
      If assigned(FOnPaintEnd) then FOnPaintEnd(self, FCanvas);
    except
    end;
  finally
   {$IFDEF WIN32}
    metaCanvas.Free;
   {$ELSE}
    Metafile.Close;
   {$ENDIF}
    try
      wmf.Assign(metafile);
    finally
     metafile.free;
     FCanvas := Canvas;
    {$IFNDEF WIN32}
     Width := saveWidth;
     Height := saveHeight;
    {$ENDIF}
    end;
  end;
end;

procedure TxyGraph.FillBitmap(bmp:TBitmap);
var iBmp:TBitmap;
begin
  iBmp := TBitmap.create;
  ibmp.width := bmp.width;
  ibmp.height := bmp.height;
  FCurrWidth := bmp.width;
  FCurrHeight := bmp.height;
  FCanvas:= iBmp.Canvas;
  FAltCanvas := iBmp.Canvas;
  try
    PaintGraph;
    try
      If assigned(FOnPaintEnd) then FOnPaintEnd(self, FCanvas);
    except
    end;
    bmp.Assign(iBmp);
  finally
    FCanvas := Canvas;
    FAltCanvas := Canvas;
  end;
end;


{---------------------------------------------------------------------
    #6e. TxyGraph  - property servers
 ---------------------------------------------------------------------}

procedure TxyGraph.SetPlotting(v: Boolean);
begin
  FPlotting := v;
  Paint;
  if v then DoRescaleEvent;
end;

procedure TxyGraph.SetHasDragged(v:boolean);
{ this procedure is primarily present to set FHasdragged to false
  and override any user changes }
begin
 if v <> FHasDragged then
  if v then
  begin
    FHasDragged := true;
    FHoldXAS := FXAxis.Autosizing;
    FHoldXmin := FXAxis.Min;
    FHoldXmax := FXAxis.Max;
    FHoldYAS := FYAxis.Autosizing;
    FHoldYmin := FYAxis.Min;
    FHoldYmax := FYAxis.Max;
    FHoldY2AS := FYAxis2.Autosizing;
    FHoldY2min := FYAxis2.Min;
    FHoldY2max := FYAxis2.Max;
  end
  else
  begin
    FHasDragged := false;
    plotting := false;
    FXaxis.autosizing := FHoldXAS;
    if not FHoldXAS then
    begin
      FXAxis.min := FHoldXMin;
      FXAxis.max := FHoldXMax;
    end;
    FYaxis.autosizing := FHoldYAS;
    if not FHoldYAS then
    begin
      FYAxis.min := FHoldYMin;
      FYAxis.max := FHoldYMax;
    end;
    FYaxis2.autosizing := FHoldY2AS;
    if not FHoldY2AS then
    begin
      FYAxis2.min := FHoldY2Min;
      FYAxis2.max := FHoldY2Max;
    end;
    plotting := true;
  end;
end;

{---------------------------------------------------------------------
    #6f. TxyGraph  - series routines
 ---------------------------------------------------------------------}

function TxyGraph.GetSeriesInt;
var p: TSeries;
begin
  p := FSeries;
  while (p <> nil) and (p.FSeriesIndex <> i) do p := p.next;
  if p = nil then
    result := CreateNewSeries(i)
  else
    Result := p;
end;

function TxyGraph.CreateNewSeries;
var t,p,p1:TSeries;
begin
{note! the list of series is maintained on insertion here,
       but on deletion in the TSeries.destroy routine }
  t := TSeries.create(self,i);
  result := t;
  if Fseries = nil then
    Fseries := t
  else
  begin
    p := FSeries;
    p1 := nil;
    while (p <> nil) and (p.FSeriesIndex < i) do
      begin
      p1 := p;
      p := p.next;
      end;
    if p1 = nil then
      begin
      t.next := Fseries;
      Fseries := t;
      end
    else
      begin
      p1.next := t;
      t.next := p;
      end;
  end;
  if assigned(FOnNewSeries) then FOnNewSeries(self,i);
end;

{---------------------------------------------------------------------
    #6g. TxyGraph  - series hooking routines
 ---------------------------------------------------------------------}

procedure TxyGraph.Hookseries(Source:TSeries; Ahooktype:THooktype;
                     Dest:Integer; WantPersist:boolean);
var p:pHookedSeries;
begin
 p := FHookedSeries;
 while (p <> nil) and (p^.id <> Source) do p := p^.next;
 if p = nil then
   begin
   new(p);
   p^.next := FHookedSeries;
   FHookedSeries := p;
   end
 else
   if Not WantPersist then getseriesint(dest).clear;
   {clear series if we're reassigning to another one unless user says otherwise }
 p^.id := source;
 p^.Series := dest;
 p^.hooktype := AHooktype;
 p^.id.hooks[self] := HookedSeriesEvent;
 HookedSeriesEvent(p^.id, dsClearUpdates); {dsClearUpdates = anything might have changed}
end;

procedure TxyGraph.unhookseries(source:TSeries; WantPersist:boolean);
var p,p1:phookedseries;
begin
 p1 := nil;
 p := FHookedSeries;
 while (p <> nil) and (p^.id <> source) do
   begin
   p1 := p;
   p := p^.next;
   end;
 if p = nil then raise exception.create('trying to unhook a series that is not hooked');
 if p1 = nil then
   FhookedSeries := p^.next
 else
   p1^.next := p^.next;
 p^.id.hooks[self] := nil;
 if Not WantPersist then getseriesint(p^.series).clear;
 dispose(p);
 if not application.terminated then paint;
end;

procedure TxyGraph.HookedSeriesEvent(Sender:TObject; TheMessage:TDSChangeType);
var p:pHookedSeries;
  procedure readhookedseriesdata;
  var ps:TSeries;
      pp:pgpoint;
  begin
   ps := getseriesint(p^.series);
   ps.holdupdates := true;
   ps.clear;
   case p^.hooktype of
    ht_plotdata:begin
       pp := p^.id.Data;
       while pp <> nil do
         begin
         ps.add(pp^.xv,pp^.yv);
         pp := pp^.next;
         end;
       end;
   {$IFDEF STATISTICS}
    ht_plotresiduals:if p^.id.RegType in [rg_Linear, rg_passingBablok,
                  rg_quadratic, rg_RunningAverage, rg_RunningAverageproper] then
       begin
       p^.id.DoRegression;
       pp := p^.id.Data;
       while pp <> nil do
         begin
         ps.add(pp^.xv,pp^.yv-pp^.rv);
         pp := pp^.next;
         end;
       end;
   {$ENDIF}
   end;
   ps.holdupdates := false;
  end;

begin
 p := FHookedSeries;
 while (p <> nil) and (p^.id <> Sender as TSeries) do p := p^.next;
 if p = nil then raise exception.create('HookedSeriesEvent but series not properly hooked');
 case TheMessage of
  dsUnLoad:unhookseries(p^.id, false);
  dsDataChange:readhookedseriesdata;
  {$IFDEF STATISTICS}
  dsRegressionChange:if p^.hooktype = ht_plotResiduals then readhookedseriesdata;
  {$ENDIF}
  dsZeroChange:{ignore};
  dsClearUpdates:readhookedseriesdata;
  dsChangeLook:{ignore};
 end;
end;

function TxyGraph.FindSeries;
var p: TSeries;
begin
  p := FSeries;
  while (p <> nil) and (p.FSeriesIndex <> i) do p := p.next;
  result := (p <> nil);
end;

{---------------------------------------------------------------------
    #6g. TxyGraph  - data manipulation
 ---------------------------------------------------------------------}

function TxyGraph.DataFound;
var p: TSeries;
begin
  Result := False;
  p := FSeries;
  while (p <> nil) and not Result do
  begin
    Result := Result or (p.FData <> nil) and ((whichaxis = FXAxis)
           or (whichaxis = nil) or (p.FWhichYAxis = whichaxis));
    p := p.next;
  end;
end;

procedure TxyGraph.ClearAll;
var holdPlotting: Boolean;
begin
  holdPlotting := FPlotting;
  FPlotting := false;
  while FSeries <> nil do FSeries.Free;
  FPlotting := holdPlotting;
  if FPlotting then Paint;
end;

{---------------------------------------------------------------------
    #6h. TxyGraph  - Marks routines
 ---------------------------------------------------------------------}

procedure TxyGraph.ClearMarks;
var om: pMark;
begin
  while FMarks <> nil do
  begin
    om := FMarks;
    FMarks := FMarks^.next;
    Dispose(om);
  end;
  if FPlotting then Paint;
end;

procedure TxyGraph.addmark;
var tm,p,p1:pmark;
begin
 p := fmarks;
 p1 := nil;
 while (p <> nil) and (p^.id < id) do
   begin
   p1 := p;
   p := p^.next;
   end;
 if (p = nil) then
   if (p1 = nil) then
     begin {first mark to be added}
     new(p);
     p^.next := nil;
     fmarks := p;
     end
   else
     begin {append to list}
     new(p);
     p1^.next := p;
     p^.next := nil;
     end
 else
   if p^.id <> id then
     if (p1 = nil) then
       begin {prepend to list}
       new(p);
       p^.next := FMarks;
       Fmarks := p;
       end
     else
       begin {insert in list}
       new(p);
       p^.next := p1^.next;
       p1^.next := p;
       end;
  p^.id := id;
  p^.color := c;
  p^.x1 := x1;
  p^.y1 := y1;
  p^.x2 := x2;
  p^.y2 := y2;
  p^.caption := name;
  p^.marktype := marktype;
  p^.status := status;
  if fplotting then paint;
end;


function TxyGraph.DeleteMark;
var p, p1: pMark;
begin
  Result := False;
  p := FMarks;
  p1 := nil;
  while (p <> nil) and (p^.id < id) do
    begin
    p1 := p;
    p := p^.next;
    end;
  if (p=nil) or (p^.id <> id) then
    result := false
  else
    begin
    result := true;
    if p1 = nil then
      FMarks := p^.next
    else
      p1^.next := p^.next;
    dispose(p);
    if FPlotting then Paint;
    end;
end;

{---------------------------------------------------------------------
    #6i. TxyGraph  - External interaction
 ---------------------------------------------------------------------}

{copy and paste routines modelled on Delphi Developers Guide,
 Pacheco & Teixeira p 352}
procedure TxyGraph.CopyToClipboard;
var
    metafile:TMetafile;

    strm:TMemoryStream;
    p:pointer;
    h:Thandle;

    bmp:TBitmap;
    Format: Word;
    Data: THandle;
    Palette: HPALETTE;

begin
  metafile := TMetafile.create;
  try
    strm := TMemoryStream.create;
    try
      bmp := TBitmap.create;
      try
        FillMetafile(metafile);
        FillBitmap(bmp);
        savetostream(strm,true);
        h := GlobalAlloc(GMEM_Moveable, strm.size);
        try
          p := GlobalLock(h);
          try
            move(strm.memory^, p^, strm.size);
            clipboard.open;
            Clipboard.SetAsHandle(CF_xyGraphFmt, h);
            bmp.SaveToClipboardFormat(Format, Data, Palette);
            Clipboard.SetAsHandle(Format, Data);
            if Palette <> 0 then Clipboard.SetAsHandle(CF_PALETTE, Palette);
            Clipboard.assign(metafile);
            Clipboard.Close;
          finally
            GlobalUnLock(h);
          end;
        except
          Globalfree(h);
          raise;
        end;
      finally
        bmp.free;
      end;
    finally
      strm.free;
    end;
  finally
    metafile.free;
  end;
end;

procedure TxyGraph.PasteFromClipboard;
var h:Thandle;
    p:Pointer;
    strm:TMemoryStream;
begin
  h := ClipBoard.GetAsHandle(CF_xyGraphFmt);
  try
    if h = 0 then exit;
    p := GlobalLock(h);
    try
      strm := TMemoryStream.create;
      try
        strm.write(p^, GlobalSize(h));
        strm.seek(0,0);
        readfromstream(strm);
      finally
        strm.free;
      end;
    finally
      GlobalunLock(h);
    end;
  except
    GlobalFree(h);
     {not sure why this is required?? Why kill the clipboard memory?
      It's required above because in the copy routine it may not have
      been handed to the clipboard. But I'm not so sure here}
    raise;
  end;
end;

procedure TxyGraph.savetoWMF(const fname:string);
var metafile:TMetaFile;
begin
  metafile := TMetaFile.Create;
  try
    fillMetaFile(metafile);
    metafile.savetofile(fname);
  finally
    metafile.free;
  end;
end;

procedure TxyGraph.SaveToBitmap(const fname:string);
var bitmap:TBitmap;
begin
  bitmap := TBitmap.Create;
  bitmap.width := width;
  bitmap.height := height;
  try
    fillBitmap(bitmap);
    bitmap.savetofile(fname);
  finally
    bitmap.free;
  end;
end;

{
saving graph to a file/stream

  These procedures save the graph to a file. As well as creating a
  special file, these also stream series to the .dfm file so that
  series created at design time are saved when compiled.

  Because the .dfm file includes a number of properties that
  shouldn't change - parents, events, etc - a whole separate
  subsystem has been created for savetofile methods. The 2
  subsystems show marks and series procedures.

  The file format is intended to be forwards compatible(!).
  The file version number written at the start of the stream
  is different to the version of xyGraph - it will change
  when a new format is created that is not backwards
  compatible.

  use savetostream to include the same info in a stream of
  your wish.

Format: (for file - a subset is relevent for the .dfm)

    header: constants declared at the top of this unit.
              - arbitrary identifier
              - file version identifier
              - code version of writing xygraph
    sections - in order: Appearance, FDimensions, FXAxis,
                         FYAxis, FYAXis2 and FLegend
    followed by 2 blocks - series count & Marks Count
    sections - a section for each series, and then a section for
               each mark
    bytes left - 4 byte indicator of number of bytes remaining in the
                 stream - to allow future versions to add information

  Each Section consists of a series of blocks ending in a block
  with a identifier of 0. identifier 99 is reserved for a descriptor.
  The Series Section consists of 3 separate sections, a TSeries
  section, followed by a TDataSeries section, followed by the data
  for the series.

  Each block consists of an identifier (integer < 100) and a format
  identifier, written as a string (optionally with the identifier
  name appended), followed by the value in the format indicated.
  Where possible, formats will be converted when reading.

  Because with the exception of the bytes left field all information
  is written with TWriter descendants each individual piece of
  information is separated by the TWriter separator.

  WantCompact can be used to increase the amount information written
  as strings or decrease the file size.

Information not saved at all for design reasons.

  Hooked series - all hooks will be lost
  Series indexes will be changed
  Intercept info will be lost
  regression data is not saved (but will be automatically rebuilt)
  Pointwarning limit is not saved
  HoldUpdates will be ignored
  Function plotting will not be saved

Known Bugs:

  When you copy to clipboard at designtime, with no series
  in the graph. I don't know why. Even though the error is raised,
  the copy had still worked successfully and you will be

  pasting - converting the series numbers - the reason is again
  unknown. Keep trying! It generally works in the end!

}

procedure TxyGraph.Assign;
var MemStrm:TMemoryStream;
begin
  if not (source is TxyGraph) then exit;
  MemStrm := TMemoryStream.create;
  try
    (source as tXygraph).savetostream(MemStrm,true);
    MemStrm.seek(0, 0);
    self.readfromstream(MemStrm);
  finally
    MemStrm.free;
  end;
end;


procedure TxyGraph.savetofile;
var f:TFilestream;
begin
  f := Tfilestream.create(fname,fmCreate);
  try
    savetostream(f,savedata);
  finally
    f.free;
  end;
end;

procedure TxyGraph.savetostream;
var w:TxyWriter;
    t:longint;
begin
  FSaveData := savedata;
  w := TxyWriter.create(f,4096);
  try
    w.writeString(xyGraphFileSig + xyGraphFileVersion + versionstamp);
    writeAppToStream(w, FAppearance);
    WriteDimToStream(w, FDimensions);
    WriteAxsToStream(w, FXAxis);
    WriteAxsToStream(w, FYAxis);
    WriteAxsToStream(w, FYAXis2);
    WriteLegToStream(w, FLegend);
    WriteseriesCount(w);
    writeMarkCount(w);
  finally
    w.free;
  end;
  if FFilingSeriesCount > 0 then WriteAllSeries(f);
  if FFilingMarksCount > 0 then WriteAllMarks(f);
  {write number of bytes left - for future compatibility}
  t := 0;
  f.write(t,4);
end;

procedure TxyGraph.readfromfile;
var f:TFilestream;
begin
  f := Tfilestream.create(fname,fmOpenRead);
  try
    readfromstream(f);
  finally
    f.free;
  end;
end;

procedure TxyGraph.readfromstream;
var r:TxyReader;
    s:string;
    offset,t:longint;
    op:boolean;
begin
  op := plotting;
  plotting := false;
  try
    clearall;
    clearmarks;
    setdefaultpropertyValues;

    r := TxyReader.create(f,4096);
    try
      s := r.readstring;
      if copy(s,1,length(xyGraphFileSig)) <> xyGraphFileSig then
        raise exception.create('Not a valid graph file');
      if s[length(xyGraphFileSig)+1] <> xyGraphFileVersion then
        raise exception.create('Wrong Graph version');
      ReadAppFromStream(r, FAppearance);
      ReadDimFromStream(r, FDimensions);
      ReadAxsFromStream(r, FXAxis);
      ReadAxsFromStream(r, FYAxis);
      ReadAxsFromStream(r, FYAXis2);
      ReadLegFromStream(r, FLegend);
      readseriesCount(r);
      readMarkCount(r);
    finally
      r.free;
    end;
    if FFilingSeriesCount > 0 then readAllSeries(f);
    if FFilingMarksCount > 0 then readAllMarks(f);
    {room for future expansion}
    f.read(t,4);
    f.position := f.position + t;
  finally
    plotting := op;
  end;
end;

procedure TxyGraph.DefineProperties;
begin
  inherited defineproperties(filer);
  Filer.DefineProperty('SeriesCount', ReadSeriesCount, WriteseriesCount, true);
  Filer.DefineProperty('MarkCount', ReadMarkCount, writeMarkCount, true);
  Filer.DefineBinaryProperty('SeriesInfo', ReadAllSeries, WriteAllSeries, FFilingSeriesCount > 0);
  Filer.DefineBinaryProperty('MarksInfo', ReadAllMarks, WriteAllMarks, FFilingMarksCount > 0);
end;

procedure TxyGraph.WriteSeriesCount;
var i:integer;
    p:TSeries;
begin
  p := Fseries;
  i := 0;
  while p <> nil do
    begin
    inc(i);
    p := p.next;
    end;
  FFilingSeriesCount := i;
  if wantcompactfile then
    begin
    writer.writestring(ktInteger);
    writer.writeinteger(FFilingSeriesCount);
    end
  else
    begin
    writer.writestring(ktstring);
    writer.writestring(inttostr(FFilingSeriesCount));
    end;
end;

procedure TxyGraph.ReadSeriesCount;
begin
  if reader.readstring = ktInteger then
    FFilingSeriesCount := reader.readinteger
  else
    FFilingSeriesCount := strtoint(reader.readstring);
end;

procedure TxyGraph.WriteAllSeries;
var p:TSeries;
    w:TxyWriter;
begin
  if FFilingSeriesCount = 0 then raise exception.create('internal filing error: TxyGraph');
  w := TxyWriter.create(stream, 4096);
  try
    p := Fseries;
    while p <> nil do
      begin
      WriteSerToStream(w, p, FSaveData);
      p := p.next;
      end;
  finally
    w.free;
  end;
end;

procedure TxyGraph.ReadAllSeries;
var p:TSeries;
    i:integer;
    r:TxyReader;
begin
  if FFilingSeriesCount = 0 then raise exception.create('internal filing error: TxyGraph');
  r := TxyReader.create(stream, 4096);
  try
    for i := 1 to FFilingSeriesCount do
      begin
      p := CreateNewSeries(i);
      ReadSerFromStream(r, p);
      end;
  finally
    r.free;
  end;
end;

procedure TxyGraph.writeMarkCount;
var i:integer;
    p:pMark;
begin
  p := FMarks;
  i := 0;
  while p <> nil do
    begin
    inc(i);
    p := p^.next;
    end;
  FFilingMarksCount := i;
  if wantcompactfile then
    begin
    writer.writestring(ktInteger);
    writer.writeinteger(FFilingMarksCount);
    end
  else
    begin
    writer.writestring(ktstring);
    writer.writestring(inttostr(FFilingMarksCount));
    end;
end;

procedure TxyGraph.ReadMarkCount;
begin
  if reader.readstring = ktInteger then
    FFilingMarksCount := reader.readinteger
  else
    FFilingMarksCount := strtoint(reader.readstring);
end;

procedure TxyGraph.WriteAllMarks;
var p:pMark;
    w:TxyWriter;
begin
  if FFilingMarksCount = 0 then raise exception.create('internal filing error2: TxyGraph');
  w := TxyWriter.create(stream, 200);
  try
    p := FMarks;
    while p <> nil do
     with w,p^ do
      begin
      PutString (99, 'Section:', 'Mark');
      Putinteger( 1, 'ID ', id);
      putfloat  ( 2, 'X1Value', x1);
      putfloat  ( 3, 'Y1Value', y1);
      putfloat  ( 4, 'X2Value', x2);
      putfloat  ( 5, 'Y2Value', y2);
      putinteger( 6, 'Color', color);
      putstring ( 7, 'Caption', caption);
      putinteger( 8, 'TypeOfMark', longint(marktype));
      putboolean( 9, 'WhenDrawn', status);
      endsection;
      p := next;
      end;
  finally
     w.free;
  end;
end;

procedure TxyGraph.ReadAllMarks;
var p:pMark;
    r:TxyReader;
    I,n,mt,c:integer;
    x1,y1,x2,y2:double;
    s:string;
    st,finished:boolean;
begin
  if FFilingMarksCount = 0 then raise exception.create('internal filing error2: TxyGraph');
  r := TxyReader.create(stream, 200);
  try
   with r do
    for i := 1 to FFilingMarksCount do
     begin
     finished := false;
     while not finished do
      case readblock of
        0: finished := true;
        1: n := getinteger;
        2: x1 := getfloat;
        3: y1 := getfloat;
        4: x2 := getfloat;
        5: y2 := getfloat;
        6: c := getinteger;
        7: s := getstring;
        8: mt := getinteger;
        9: st := getboolean;
      else readunknown;
      end;
     Addmark(n, x1, y1, x2, y2, c, s, et_MarkType(mt), st);
     end;
  finally
     r.free;
  end;
end;

begin
  CF_xyGraphFmt := RegisterClipboardFormat('Internal xyGraph Format');
end.

